# -*- coding: utf-8 -*-
"""
Created on Fri Oct 20 15:51:45 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import sys
import pandas as pd
import os

# Pyseawave toolbox for second-order waves
from pyseawave.spectrum import WaveSpectrum

# Wave spectrum toolbox and distribution scripts
sys.path.append('functions')
from functions.wave_spectra import make_spec_jons as jons
from functions.FFT_scripts import transform_spec2tt
from functions.distribution_scripts import calc_zup

# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Folder name in which to store files and figures (rel. to present folder). Note
# that the folder 'waveinputfiles' in the repository already contains the input
# files used for the paper. If you set filefolder to the same name, these files 
# will be overwritten with new wave realisations. This means that the results may 
# not be identical to the paper anymore (different wave realisation, so slightly
# different results).
filefolder = 'waveinputfiles_new'
figfolder = 'out_figs'

# Target exposure duration, typically between 0.5 and 3 hours [s]
dur = 3600

# Number of seeds, each with duration 'dur'
Ns = 50

# Wave properties for primary input spectrum
Hs = 10
Tp = 11
gamma = 3.3
omstep = 1/dur*2*np.pi
om = np.arange(0,200,omstep) # Seeds with duration <dur>
wd = 30 # water depth
Vs = 0 # current speed or vessel speed
T1 = Tp/1.198 # estimated from relation for JONSWAP

# Lowest and highest freq considered in 1st and 2nd order calc [rad/s]
# Why also in the 1st order wave calculation?
flo1 = 0
fhi1 = 5

# Define noise spectrum properties MediumInd indicator (extra wave spectrum)
Hs_s = 1.5; Tp_s = 9; gam_s = 3.3 # Noise ShortTp

# Define noise spectrum properties BadInd indicator (extra wave spectrum)
Hs_l = 2; Tp_l = 14; gam_l = 1 # Noise LongTp

# Noise texts (figure titles and names)
noisetxt_GoodInd = 'GoodInd'
noisetxt_PerfInd = 'PerfInd'
noisetxt_MediumInd = 'MediumInd'
noisetxt_BadInd = 'BadInd'

# Seed initialisation (to make sure we can do this multiple times)
# Integer number. Note that seedinit = 0 is the original result, where 
# 'real{seedinit}' is not included in the file names! For the paper we used
# 0,1,...,9. For each realisation, the script needs to be run again.
seedinit = 9



#############
# END INPUT
#############

# Check wether output directories exist, if not make them
isExist1 = os.path.exists(filefolder)
if not isExist1:
   os.makedirs(filefolder)
isExist2 = os.path.exists(figfolder)
if not isExist2:
   os.makedirs(figfolder)

# Average number of peaks in target duration
ne_exp = dur/Tp

# Exposure duration probability level in the DNR
pexp = 1/ne_exp

# Generate Jonswap spectral densities of primary spectrum
Sjons = jons(Hs,Tp,gamma,om)
    
# Initialise
EM_lin_all_PerfInd = np.zeros((Ns,))
EM_lin_all_GoodInd = np.zeros((Ns,))
EM_lin_all_BadInd = np.zeros((Ns,))
EM_lin_all_MediumInd = np.zeros((Ns,))

EM_2nd_all_scr_PerfInd = np.zeros((Ns,))
EM_2nd_all_scr_GoodInd = np.zeros((Ns,))
EM_2nd_all_scr_BadInd = np.zeros((Ns,))
EM_2nd_all_scr_MediumInd = np.zeros((Ns,))

EM_2nd_all_true = np.zeros((Ns,))

t_peaks_lin_all2_PerfInd = []
t_peaks_lin_all2_GoodInd = []
t_peaks_lin_all2_BadInd = []
t_peaks_lin_all2_MediumInd = []

t_peaks_2nd_all2_scr_PerfInd = []
t_peaks_2nd_all2_scr_GoodInd = []
t_peaks_2nd_all2_scr_BadInd = []
t_peaks_2nd_all2_scr_MediumInd = []

t_peaks_2nd_all2_true = []

peaks_lin_all2_PerfInd = []
peaks_lin_all2_GoodInd = []
peaks_lin_all2_BadInd = []
peaks_lin_all2_MediumInd = []

peaks_2nd_all2_scr_PerfInd = []
peaks_2nd_all2_scr_GoodInd = []
peaks_2nd_all2_scr_BadInd = []
peaks_2nd_all2_scr_MediumInd = []

peaks_2nd_all2_true = []

# Loop over N seeds
for N in range(Ns):
    
    print(f'N={N}')
    
    # Generate corresponding random phase array for the new seed
    np.random.seed(seedinit*10000+N)
    eps = np.random.rand(len(Sjons))*2*np.pi

    # Generate Pyseawave spectrum object
    wave_psw_spec = WaveSpectrum(freq=om,psd=Sjons,phase=eps,depth=wd,freq_lo1=flo1,freq_hi1=fhi1,freq_lo2=flo1,freq_hi2=fhi1)
   
    # Linear and second-order wave time traces (+total)
    time = wave_psw_spec.time()
    wave_lin = wave_psw_spec.linear_elev()
    sum_freq_elev = wave_psw_spec.sum_freq_elev()
    dif_freq_elev = wave_psw_spec.dif_freq_elev()
    wave_2nd = wave_lin + sum_freq_elev + dif_freq_elev
    
    # Plot time histories linear and 2nd order waves
    if N == 0:
        # plt.figure(2,figsize=(10,4))
        plt.figure(2,figsize=(7,5))
        plt.plot(time, wave_2nd, label='Total nonlinear wave')
        plt.plot(time, wave_lin, label='Linear wave')
        plt.plot(time, sum_freq_elev, label='Sum-freq. elev.')
        plt.plot(time, dif_freq_elev, label='Diff-freq. elev.')
        plt.xlabel('t [s]')
        plt.ylabel('$\eta$ [m]')
        plt.legend()
        plt.xlim([0,np.max(time)])
        plt.savefig(f'{figfolder}/Lin2nd_TimeHistories_{noisetxt_GoodInd}_real{seedinit}_seed{N}.pdf',bbox_inches='tight')

        # Zoom in around max crest and save again
        i_cmax = np.argmax(wave_2nd)
        plt.xlim([time[i_cmax]-50,time[i_cmax]+50])
        plt.savefig(f'{figfolder}/Lin2nd_TimeHistories_{noisetxt_GoodInd}_real{seedinit}_seed{N}_ZOOM.pdf',bbox_inches='tight')
    
    # Define LongTp and ShortTp noise 
        
    # JONSWAP spectrum with LongTp noise
    Sjons_l = jons(Hs_l,Tp_l,gam_l,om)
    seed_l = seedinit*10000+4000+N
    time_l,wave_lin_l,eps_l = transform_spec2tt(om,Sjons_l,seed_l)
    
    # JONSWAP spectrum with ShortTp noise
    Sjons_s = jons(Hs_s,Tp_s,gam_s,om)
    # eps_s = np.random.rand(len(Sjons_s))*2*np.pi
    seed_s = seedinit*10000+4000+N
    time_s,wave_lin_s,eps_s = transform_spec2tt(om,Sjons_s,seed_s)

    # Linear wave data for all four noise options (second-order is identical for all)
    wave_lin_PerfInd = wave_2nd.copy() # Use 2nd order also as linear signal to check perfect indicator case
    wave_lin_GoodInd = wave_lin.copy()
    wave_lin_BadInd = wave_lin + wave_lin_l
    wave_lin_MediumInd = wave_lin + wave_lin_s

    ## Apply screening procedure based on linear wave elevation
    
    # Peak identification linear time traces (used as true linear data & for screening)
    t_peaks_lin_PerfInd,peaks_lin_PerfInd,_,_,t_zup_lin_PerfInd = calc_zup(time,wave_lin_PerfInd,ploton=0)
    t_peaks_lin_GoodInd,peaks_lin_GoodInd,_,_,t_zup_lin_GoodInd = calc_zup(time,wave_lin_GoodInd,ploton=0)
    t_peaks_lin_BadInd,peaks_lin_BadInd,_,_,t_zup_lin_BadInd = calc_zup(time,wave_lin_BadInd,ploton=0)
    t_peaks_lin_MediumInd,peaks_lin_MediumInd,_,_,t_zup_lin_MediumInd = calc_zup(time,wave_lin_MediumInd,ploton=0)
    
    # Peak identification 2nd order time traces (used as true 2nd order data)
    t_peaks_2nd_true,peaks_2nd_true,_,_,t_zup_2nd_true = calc_zup(time,wave_2nd,ploton=0)
    
    # Find matching max values between indicator zero up-crossings (to match in time)
    t_peaks_2nd_scr_PerfInd = np.zeros(np.shape(peaks_lin_PerfInd))*np.nan
    peaks_2nd_scr_PerfInd = np.zeros(np.shape(peaks_lin_PerfInd))*np.nan
    t_peaks_2nd_scr_GoodInd = np.zeros(np.shape(peaks_lin_GoodInd))*np.nan
    peaks_2nd_scr_GoodInd = np.zeros(np.shape(peaks_lin_GoodInd))*np.nan
    t_peaks_2nd_scr_BadInd = np.zeros(np.shape(peaks_lin_BadInd))*np.nan
    peaks_2nd_scr_BadInd = np.zeros(np.shape(peaks_lin_BadInd))*np.nan
    t_peaks_2nd_scr_MediumInd = np.zeros(np.shape(peaks_lin_MediumInd))*np.nan
    peaks_2nd_scr_MediumInd = np.zeros(np.shape(peaks_lin_MediumInd))*np.nan
    for ii, dum in enumerate(peaks_lin_PerfInd[0:-1]):
        split = wave_2nd[(time>t_zup_lin_PerfInd[ii]) & (time<t_zup_lin_PerfInd[ii+1])]
        t_split = time[(time>t_zup_lin_PerfInd[ii]) & (time<t_zup_lin_PerfInd[ii+1])]
        i_split_p = np.argmax(split)
        t_peaks_2nd_scr_PerfInd[ii] = t_split[i_split_p]
        peaks_2nd_scr_PerfInd[ii] = split[i_split_p]
    for ii, dum in enumerate(peaks_lin_GoodInd[0:-1]):
        split = wave_2nd[(time>t_zup_lin_GoodInd[ii]) & (time<t_zup_lin_GoodInd[ii+1])]
        t_split = time[(time>t_zup_lin_GoodInd[ii]) & (time<t_zup_lin_GoodInd[ii+1])]
        i_split_p = np.argmax(split)
        t_peaks_2nd_scr_GoodInd[ii] = t_split[i_split_p]
        peaks_2nd_scr_GoodInd[ii] = split[i_split_p]
    for ii, dum in enumerate(peaks_lin_BadInd[0:-1]):
        split = wave_2nd[(time>t_zup_lin_BadInd[ii]) & (time<t_zup_lin_BadInd[ii+1])]
        t_split = time[(time>t_zup_lin_BadInd[ii]) & (time<t_zup_lin_BadInd[ii+1])]
        i_split_p = np.argmax(split)
        t_peaks_2nd_scr_BadInd[ii] = t_split[i_split_p]
        peaks_2nd_scr_BadInd[ii] = split[i_split_p]
    for ii, dum in enumerate(peaks_lin_MediumInd[0:-1]):
        split = wave_2nd[(time>t_zup_lin_MediumInd[ii]) & (time<t_zup_lin_MediumInd[ii+1])]
        t_split = time[(time>t_zup_lin_MediumInd[ii]) & (time<t_zup_lin_MediumInd[ii+1])]
        i_split_p = np.argmax(split)
        t_peaks_2nd_scr_MediumInd[ii] = t_split[i_split_p]
        peaks_2nd_scr_MediumInd[ii] = split[i_split_p]

    # Time trace plot with peaks
    if N == 1:
        # fig3 = plt.figure(3,figsize=(10,4))
        fig3 = plt.figure(3,figsize=(7,5))
        ax3 = fig3.add_subplot(1,1,1)
        pl1 = ax3.plot(time,wave_lin,'-',color='grey',label='linear')
        pl2 = ax3.plot(time,wave_2nd,'--',color='red',label='2nd order')
        pl3 = ax3.plot(t_zup_lin_GoodInd,np.zeros(np.shape(t_zup_lin_GoodInd)),marker='+',color='black',label='linear zero up-crossings',linestyle = 'None',markersize=9)
        pl4 = ax3.plot(t_peaks_lin_GoodInd,peaks_lin_GoodInd,marker='o',color='grey',label='linear peaks',linestyle = 'None')
        pl5 = ax3.plot(t_peaks_2nd_scr_GoodInd,peaks_2nd_scr_GoodInd,marker='d',color='red',label='corresponding 2nd order peaks',linestyle = 'None')
        plt.title('Identification of zero up-crossings, peaks and troughs')
        plt.legend(handles=[pl1[0],pl2[0],pl3[0],pl4[0],pl5[0]])
        plt.xlabel('t [s]')
        plt.ylabel('$\eta$ [m]')
        plt.savefig(f'{figfolder}/Fig4_Lin2nd_TimeHistories_{noisetxt_GoodInd}_CrestMatch_real{seedinit}_seed{N}.pdf',bbox_inches='tight')
  
        # Zoom in around max crest and save again
        plt.xlim([time[i_cmax]-50,time[i_cmax]+50])
        plt.savefig(f'{figfolder}/Fig4_Lin2nd_TimeHistories_{noisetxt_GoodInd}_CrestMatch_real{seedinit}_seed{N}_ZOOM.pdf',bbox_inches='tight')
 
    # Save ensemble maxima results for the present seed (true and from screening)
    EM_lin_all_PerfInd[N] = max(peaks_lin_PerfInd)
    EM_lin_all_GoodInd[N] = max(peaks_lin_GoodInd)
    EM_lin_all_BadInd[N] = max(peaks_lin_BadInd)
    EM_lin_all_MediumInd[N] = max(peaks_lin_MediumInd)
    EM_2nd_all_scr_PerfInd[N] = max(peaks_2nd_scr_PerfInd)
    EM_2nd_all_scr_GoodInd[N] = max(peaks_2nd_scr_GoodInd)
    EM_2nd_all_scr_BadInd[N] = max(peaks_2nd_scr_BadInd)
    EM_2nd_all_scr_MediumInd[N] = max(peaks_2nd_scr_MediumInd)
    EM_2nd_all_true[N] = max(peaks_2nd_true)
    
    # Save data for all seeds
    if N > 0:
        t_peaks_lin_all2_PerfInd = np.append(t_peaks_lin_all2_PerfInd,t_peaks_lin_PerfInd+max(t_peaks_lin_all2_PerfInd))
        t_peaks_lin_all2_GoodInd = np.append(t_peaks_lin_all2_GoodInd,t_peaks_lin_GoodInd+max(t_peaks_lin_all2_GoodInd))
        t_peaks_lin_all2_BadInd = np.append(t_peaks_lin_all2_BadInd,t_peaks_lin_BadInd+max(t_peaks_lin_all2_BadInd))
        t_peaks_lin_all2_MediumInd = np.append(t_peaks_lin_all2_MediumInd,t_peaks_lin_MediumInd+max(t_peaks_lin_all2_MediumInd))
        
        t_peaks_2nd_all2_scr_PerfInd = np.append(t_peaks_2nd_all2_scr_PerfInd,t_peaks_2nd_scr_PerfInd+max(t_peaks_2nd_all2_scr_PerfInd))
        t_peaks_2nd_all2_scr_GoodInd = np.append(t_peaks_2nd_all2_scr_GoodInd,t_peaks_2nd_scr_GoodInd+max(t_peaks_2nd_all2_scr_GoodInd))
        t_peaks_2nd_all2_scr_BadInd = np.append(t_peaks_2nd_all2_scr_BadInd,t_peaks_2nd_scr_BadInd+max(t_peaks_2nd_all2_scr_BadInd))
        t_peaks_2nd_all2_scr_MediumInd = np.append(t_peaks_2nd_all2_scr_MediumInd,t_peaks_2nd_scr_MediumInd+max(t_peaks_2nd_all2_scr_MediumInd))
        
        t_peaks_2nd_all2_true = np.append(t_peaks_2nd_all2_true,t_peaks_2nd_true+max(t_peaks_2nd_all2_true))
    else:
        t_peaks_lin_all2_PerfInd = np.append(t_peaks_lin_all2_PerfInd,t_peaks_lin_PerfInd)
        t_peaks_lin_all2_GoodInd = np.append(t_peaks_lin_all2_GoodInd,t_peaks_lin_GoodInd)
        t_peaks_lin_all2_BadInd = np.append(t_peaks_lin_all2_BadInd,t_peaks_lin_BadInd)
        t_peaks_lin_all2_MediumInd = np.append(t_peaks_lin_all2_MediumInd,t_peaks_lin_MediumInd)
        
        t_peaks_2nd_all2_scr_PerfInd = np.append(t_peaks_2nd_all2_scr_PerfInd,t_peaks_2nd_scr_PerfInd)
        t_peaks_2nd_all2_scr_GoodInd = np.append(t_peaks_2nd_all2_scr_GoodInd,t_peaks_2nd_scr_GoodInd)
        t_peaks_2nd_all2_scr_BadInd = np.append(t_peaks_2nd_all2_scr_BadInd,t_peaks_2nd_scr_BadInd)
        t_peaks_2nd_all2_scr_MediumInd = np.append(t_peaks_2nd_all2_scr_MediumInd,t_peaks_2nd_scr_MediumInd)
        
        t_peaks_2nd_all2_true = np.append(t_peaks_2nd_all2_true,t_peaks_2nd_true)
        
    peaks_lin_all2_PerfInd = np.append(peaks_lin_all2_PerfInd,peaks_lin_PerfInd)
    peaks_lin_all2_GoodInd = np.append(peaks_lin_all2_GoodInd,peaks_lin_GoodInd)
    peaks_lin_all2_BadInd = np.append(peaks_lin_all2_BadInd,peaks_lin_BadInd)
    peaks_lin_all2_MediumInd = np.append(peaks_lin_all2_MediumInd,peaks_lin_MediumInd)
    
    peaks_2nd_all2_scr_PerfInd = np.append(peaks_2nd_all2_scr_PerfInd,peaks_2nd_scr_PerfInd)
    peaks_2nd_all2_scr_GoodInd = np.append(peaks_2nd_all2_scr_GoodInd,peaks_2nd_scr_GoodInd)
    peaks_2nd_all2_scr_BadInd = np.append(peaks_2nd_all2_scr_BadInd,peaks_2nd_scr_BadInd)
    peaks_2nd_all2_scr_MediumInd = np.append(peaks_2nd_all2_scr_MediumInd,peaks_2nd_scr_MediumInd)
    
    peaks_2nd_all2_true = np.append(peaks_2nd_all2_true,peaks_2nd_true)
  
    # Remove data
    del t_peaks_lin_PerfInd,t_peaks_2nd_scr_PerfInd,peaks_lin_PerfInd,peaks_2nd_scr_PerfInd
    del t_peaks_lin_GoodInd,t_peaks_2nd_scr_GoodInd,peaks_lin_GoodInd,peaks_2nd_scr_GoodInd
    del t_peaks_lin_BadInd,t_peaks_2nd_scr_BadInd,peaks_lin_BadInd,peaks_2nd_scr_BadInd
    del t_peaks_lin_MediumInd,t_peaks_2nd_scr_MediumInd,peaks_lin_MediumInd,peaks_2nd_scr_MediumInd
    del t_peaks_2nd_true,peaks_2nd_true
    

# Remove NaN values
nanmask1a = ~np.isnan(t_peaks_lin_all2_PerfInd)
peaks_lin_all_PerfInd = peaks_lin_all2_PerfInd[nanmask1a]
t_peaks_lin_all_PerfInd = t_peaks_lin_all2_PerfInd[nanmask1a]
peaks_2nd_all_scr_PerfInd = peaks_2nd_all2_scr_PerfInd[nanmask1a]
t_peaks_2nd_all_scr_PerfInd = t_peaks_2nd_all2_scr_PerfInd[nanmask1a]

nanmask1b = ~np.isnan(t_peaks_lin_all2_GoodInd)
peaks_lin_all_GoodInd = peaks_lin_all2_GoodInd[nanmask1b]
t_peaks_lin_all_GoodInd = t_peaks_lin_all2_GoodInd[nanmask1b]
peaks_2nd_all_scr_GoodInd = peaks_2nd_all2_scr_GoodInd[nanmask1b]
t_peaks_2nd_all_scr_GoodInd = t_peaks_2nd_all2_scr_GoodInd[nanmask1b]

nanmask1c = ~np.isnan(t_peaks_lin_all2_BadInd)
peaks_lin_all_BadInd = peaks_lin_all2_BadInd[nanmask1c]
t_peaks_lin_all_BadInd = t_peaks_lin_all2_BadInd[nanmask1c]
peaks_2nd_all_scr_BadInd = peaks_2nd_all2_scr_BadInd[nanmask1c]
t_peaks_2nd_all_scr_BadInd = t_peaks_2nd_all2_scr_BadInd[nanmask1c]

nanmask1d = ~np.isnan(t_peaks_lin_all2_MediumInd)
peaks_lin_all_MediumInd = peaks_lin_all2_MediumInd[nanmask1d]
t_peaks_lin_all_MediumInd = t_peaks_lin_all2_MediumInd[nanmask1d]
peaks_2nd_all_scr_MediumInd = peaks_2nd_all2_scr_MediumInd[nanmask1d]
t_peaks_2nd_all_scr_MediumInd = t_peaks_2nd_all2_scr_MediumInd[nanmask1d]

nanmask2 = ~np.isnan(t_peaks_2nd_all2_true)
peaks_2nd_all_true = peaks_2nd_all2_true[nanmask2]
t_peaks_2nd_all_true = t_peaks_2nd_all2_true[nanmask2]

# # Check concatenation of all peaks
# fig4 = plt.figure(4,figsize=(10,4))
# ax4 = fig4.add_subplot(1,1,1)
# pl4 = ax4.plot(t_peaks_lin_all,peaks_lin_all,marker='o',color='blue',label='all linear peaks',linestyle = 'None')
# pl5 = ax4.plot(t_peaks_2nd_all_scr,peaks_2nd_all_scr,marker='d',color='red',label='all corresponding 2nd order peaks',linestyle = 'None')
# plt.title('All peaks')
# plt.legend(handles=[pl4[0],pl5[0]])
# plt.xlabel('t [s]')
# plt.ylabel('$\eta$ [m]')

# Write CSV files using pandas - TRUE HF DATA
AllTrueDat_true_PerfInd = np.transpose([t_peaks_2nd_all_true,peaks_2nd_all_true])
my_df1b_PerfInd = pd.DataFrame(AllTrueDat_true_PerfInd)
my_df1b_PerfInd.to_csv(f'{filefolder}/Lin2nd_trueHFdata_real{seedinit}_{Ns}seeds.csv',header = False, index= False)


# Write CSV files using pandas - PERFECT INDICATOR
AllTrueDat_scr_PerfInd = np.transpose([t_peaks_lin_all_PerfInd,peaks_lin_all_PerfInd,t_peaks_2nd_all_scr_PerfInd,peaks_2nd_all_scr_PerfInd])
my_df1a_PerfInd = pd.DataFrame(AllTrueDat_scr_PerfInd)
my_df1a_PerfInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdata_{noisetxt_PerfInd}_real{seedinit}_{Ns}seeds_scr.csv',header = False, index= False)

AllTrueEM_PerfInd = np.transpose([np.transpose(EM_lin_all_PerfInd),np.transpose(EM_2nd_all_scr_PerfInd),np.transpose(EM_2nd_all_true)])
my_df2_PerfInd = pd.DataFrame(AllTrueEM_PerfInd)
my_df2_PerfInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdataEM_{noisetxt_PerfInd}_real{seedinit}_{Ns}seeds.csv',header = False, index= False)


# Write CSV files using pandas - GOOD INDICATOR
AllTrueDat_scr_GoodInd = np.transpose([t_peaks_lin_all_GoodInd,peaks_lin_all_GoodInd,t_peaks_2nd_all_scr_GoodInd,peaks_2nd_all_scr_GoodInd])
my_df1a_GoodInd = pd.DataFrame(AllTrueDat_scr_GoodInd)
my_df1a_GoodInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdata_{noisetxt_GoodInd}_real{seedinit}_{Ns}seeds_scr.csv',header = False, index= False)

AllTrueEM_GoodInd = np.transpose([np.transpose(EM_lin_all_GoodInd),np.transpose(EM_2nd_all_scr_GoodInd),np.transpose(EM_2nd_all_true)])
my_df2_GoodInd = pd.DataFrame(AllTrueEM_GoodInd)
my_df2_GoodInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdataEM_{noisetxt_GoodInd}_real{seedinit}_{Ns}seeds.csv',header = False, index= False)


# Write CSV files using pandas - MEDIUM INDICATOR
AllTrueDat_scr_MediumInd = np.transpose([t_peaks_lin_all_MediumInd,peaks_lin_all_MediumInd,t_peaks_2nd_all_scr_MediumInd,peaks_2nd_all_scr_MediumInd])
my_df1a_MediumInd = pd.DataFrame(AllTrueDat_scr_MediumInd)
my_df1a_MediumInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdata_{noisetxt_MediumInd}_real{seedinit}_{Ns}seeds_scr.csv',header = False, index= False)

AllTrueEM_MediumInd = np.transpose([np.transpose(EM_lin_all_MediumInd),np.transpose(EM_2nd_all_scr_MediumInd),np.transpose(EM_2nd_all_true)])
my_df2_MediumInd = pd.DataFrame(AllTrueEM_MediumInd)
my_df2_MediumInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdataEM_{noisetxt_MediumInd}_real{seedinit}_{Ns}seeds.csv',header = False, index= False)


# Write CSV files using pandas - BAD INDICATOR
AllTrueDat_scr_BadInd = np.transpose([t_peaks_lin_all_BadInd,peaks_lin_all_BadInd,t_peaks_2nd_all_scr_BadInd,peaks_2nd_all_scr_BadInd])
my_df1a_BadInd = pd.DataFrame(AllTrueDat_scr_BadInd)
my_df1a_BadInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdata_{noisetxt_BadInd}_real{seedinit}_{Ns}seeds_scr.csv',header = False, index= False)

AllTrueEM_BadInd = np.transpose([np.transpose(EM_lin_all_BadInd),np.transpose(EM_2nd_all_scr_BadInd),np.transpose(EM_2nd_all_true)])
my_df2_BadInd = pd.DataFrame(AllTrueEM_BadInd)
my_df2_BadInd.to_csv(f'{filefolder}/Lin2nd_trueLFandHFdataEM_{noisetxt_BadInd}_real{seedinit}_{Ns}seeds.csv',header = False, index= False)
