# -*- coding: utf-8 -*-
"""
Created on 19 Oct 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import sys 
import os

# Distributions (Rayleigh and Forristall), first add path to functions
sys.path.append('functions')
from functions.theoretical_distributions import make_cdf_Rayleigh as Rayl
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.distribution_scripts import calc_distribution,MPM_from_DNR

# Allow for subsampling of large datasets for plotting & saving as PDF
from functions.subsample_routines import subsample_binned

# Close all figures
plt.close('all')


#############
# START INPUT
#############
           
# Figure sizes (units unclear, see matplotlib doc)
fsize_h = 7
fsize_v = 5
  
# Indicator quality case (to load the right case study dataset). Options 'PerfInd',
# 'GoodInd', 'MediumInd' and 'BadInd'.
# indcase = 'PerfInd'
indcase = 'GoodInd'
# indcase = 'MediumInd'
# indcase = 'BadInd'

# Folder name where input files are stored (files to read)
inputfolder = 'waveinputfiles'

# Folder name in which to store files and figures (rel. to present folder)
figfolder = 'out_figs'
      
# Number of seeds (to load the right case study dataset), and the number of 
# 'extra' realisations with this duration. Needs to match the files in folder 
# 'inputfolder'. Paper uses Ns = 50, Nrealisextra = 9.
Ns = 50    
Nrealisextra = 9

# Target duration [s]
dur = 3600

# Wave properties basic wave system in all cases
Hs = 10
Tp = 11
# gamma = 3.3
wd = 30 # water depth
Vs = 0 # current speed or vessel speed
T1 = Tp/1.198 # estimated from relation for JONSWAP

# Input settings Rayleigh and Forristall, range of values
VrangeRF = np.linspace(0, 15, 101)[:, None]
   

#############
# END INPUT
#############

# Check wether output directories exist, if not make them
isExist = os.path.exists(figfolder)
if not isExist:
   os.makedirs(figfolder)

# Total available duration [s]
dexp = Ns*dur
   
# Derive target exposure duration probability level
ne_exp = dur/Tp
pexp = 1/ne_exp

# Define ref LF Rayleigh and HF Forristall input functions
p_Ray,dum1 = Rayl(VrangeRF,Hs)
p_For,dum2 = Forr(VrangeRF,Hs,T1,wd,Vs)
if indcase == 'GoodInd':
    MPM_true_For = MPM_from_DNR(VrangeRF,p_For,ne_exp)
else:
    MPM_true_For = np.nan

# Read original MCS linear and 2nd order true wave time traces from CSV file
AllTrueDat_scr = np.loadtxt(f'{inputfolder}/Lin2nd_trueLFandHFdata_{indcase}_{Ns}seeds_scr.csv', delimiter=',')
t_peaks_lin_all = AllTrueDat_scr[:,0]
peaks_lin_all = AllTrueDat_scr[:,1]
t_peaks_2nd_all_scr = AllTrueDat_scr[:,2]
peaks_2nd_all_scr = AllTrueDat_scr[:,3]
AllTrueDat_true = np.loadtxt(f'{inputfolder}/Lin2nd_trueLFandHFdata_{indcase}_{Ns}seeds_true2nd.csv', delimiter=',')
t_peaks_2nd_all_true = AllTrueDat_true[:,0]
peaks_2nd_all_true = AllTrueDat_true[:,1]

print('Reading CSV with MCS over seeds done')

# Calculate true distributions for 'Monte-Carlo' dataset (all data, validation)
v_true_l,p_true_l = calc_distribution(peaks_lin_all)
v_true_h,p_true_h = calc_distribution(peaks_2nd_all_true)

# Calculate / derive true MPM from DNR's
MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp)

# Subsample the MCS true distribution data to avoid large PDF files
binsize = 0.1
bins = np.arange(0,15,binsize)
v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,bins)
v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,bins)

# Initialise plots 
fig1 = plt.figure(1,figsize=(fsize_h,fsize_v))
ax1 = fig1.add_subplot(1,1,1)
ax1.set_yscale('log')
l, b, w, h = ax1.get_position().bounds
ax1.set_position([l, b, w, h*0.95]) # make less high so 3 line title fits

# Define colors
colMCS_LF = 'black'
colMCS_HF = 'red'
colrealLF = 'grey'
colrealHF = 'pink'

# initialise
MPM_true_MCS_extra = np.zeros(Nrealisextra)

# Load and plot data for 9 alternative 50-seed realisations (to show confidence in 1-hour MPM)
for jj in range(Nrealisextra):

    ## EXTRA SEEDS: Read MCS linear and 2nd order true wave time traces from CSV file
    realisation = jj+1
    
    # True LF data and screening data - different for each noise/indicator case
    AllTrueDat_scr_R = np.loadtxt(f'{inputfolder}/Lin2nd_trueLFandHFdata_{indcase}_real{realisation}_{Ns}seeds_scr.csv', delimiter=',')
    # t_peaks_lin_all_R = AllTrueDat_scr_R[:,0]
    peaks_lin_all_R = AllTrueDat_scr_R[:,1]
    # t_peaks_2nd_all_scr_R = AllTrueDat_scr_R[:,2]
    # peaks_2nd_all_scr_R = AllTrueDat_scr_R[:,3]
    
    # True HF data - the same for all noise/indicator cases
    AllTrueDat_true_R = np.loadtxt(f'{inputfolder}/Lin2nd_trueHFdata_real{realisation}_{Ns}seeds.csv', delimiter=',')
    # t_peaks_2nd_all_true_R = AllTrueDat_true_R[:,0]
    peaks_2nd_all_true_R = AllTrueDat_true_R[:,1]
    
    # Make distributions
    v_true_l_R,p_true_l_R = calc_distribution(peaks_lin_all_R)
    v_true_h_R,p_true_h_R = calc_distribution(peaks_2nd_all_true_R)
    
    # Subsample
    v_true_l_R_sampled,p_true_l_R_sampled = subsample_binned(v_true_l_R,p_true_l_R,bins)
    v_true_h_R_sampled,p_true_h_R_sampled = subsample_binned(v_true_h_R,p_true_h_R,bins)
    
    # Add these extra realisations to the plot
    if jj == 0:
        ax1.plot(v_true_l_R_sampled,p_true_l_R_sampled,marker='.',color=colrealLF,linestyle='None',label='LF other realisations') 
        ax1.plot(v_true_h_R_sampled,p_true_h_R_sampled,marker='.',color=colrealHF,linestyle='None',label='HF other realisations')
    else:
        ax1.plot(v_true_l_R_sampled,p_true_l_R_sampled,marker='.',color=colrealLF,linestyle='None') 
        ax1.plot(v_true_h_R_sampled,p_true_h_R_sampled,marker='.',color=colrealHF,linestyle='None')
        
    # Derive MPM from extra realisations
    MPM_true_MCS_extra[jj] = MPM_from_DNR(v_true_h_R,p_true_h_R,ne_exp)

# Add true MCS DNR of the used realisation, Rayleigh / Forristall and Pexp + finish and save plot
ax1.plot(VrangeRF,p_Ray,'-.',color=[0.8,0.8,0.8],label='LF Rayleigh linear reference')
ax1.plot(VrangeRF,p_For,'--',color=[0.8,0.8,0.8],label='HF Forristall 2nd order reference')
ax1.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=colMCS_LF,linestyle='None',label='LF "true" MCS (used realisation)') 
ax1.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color=colMCS_HF,linestyle='None',label='HF "true" MCS (used realisation)')
ax1.plot([0,15],[pexp,pexp],':',color='black')
ax1.text(14,pexp*1.4,'$P_{exp}$')
plt.ylabel('P(C$\geq$c) [-]')
plt.xlabel('c [m]')
plt.ylim([10**(-5),1])
plt.xlim([0,15])
plt.legend()
handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
order = [4,5,0,1,2,3] # specify order of items in legend (and which numbers to add)
ax1.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) # add legend to plot
plt.savefig(f'{figfolder}/Fig5_MCSinput_{indcase}_incl_{Nrealisextra}_extra_{Ns}-seed_reals.pdf',bbox_inches='tight')

## Print the MPM data for MCS, including confidence interval.
MPM_true_MCS_extra_mean = np.mean(MPM_true_MCS_extra)
MPM_true_MCS_extra_U95 = 1.96*np.std(MPM_true_MCS_extra)

print(f'{dur/3600}h MPM True MCS used {Ns}-seed realisation: {MPM_true_MCS} m') # Original 50-seed realisation
print(f'{dur/3600}h MPM True MCS {Nrealisextra+1} {Ns}-seed realisations: mean {MPM_true_MCS_extra_mean} m with U95 {MPM_true_MCS_extra_U95} m') # Mean and U95% over all 50seed realisations
print(f'{dur/3600}h MPM True Forristall: {MPM_true_For} m') 

