# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import os

# Color functions
from matplotlib import cm
from matplotlib.colors import Normalize

# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Select GPR version 'GPR' for single-fidelity, or 'MF-GPR' for multi-fidelity.
# GPRvers = 'GPR'
GPRvers = 'MF-GPR'

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'B'

# Selected acquisition function for some of the figures. Note: you can only 
# select functions for which you first generated data with the Adaptive Screening 
# method. Options: 'US', 'USMV', 'USPD' or 'IVR'.
acqsel = 'USMV'

# Folder names in which the CSV files generated by Adaptive Screening were stored, 
# and name of the folder in which new figures will be stored. Defined relative 
# to the present folder, and the folder with this name needs to exist.
filefolder = 'out_files'
figfolder = 'out_figs'

# Kernel name single-fidelity HF-GPR. Note: you can only select kernels with 
# which you first generated data with the Adaptive Screening method. Preferred: 
# 'Mat32' kernel for all three.
HFkern = 'Mat32'
MFkern1 = 'Mat32'
MFkern2 = 'Mat32'

# Target exceedance duration in [s]. In the paper we used 1 hour.
dur = 3600

# Derive target probability of exceedance, using the average number of wave 
# encounters in target duration. These numbers are valid for our test case 
# study from the paper*. Only 'ne_exp' is used in the remainder, peak wave 
# period Tp is not.
Tp = 11
ne_exp = dur/Tp 

# Input: true HF MPM value from the MCS. Obtained using the script 'Generate_Lin_2nd_MCS_data.py'
MPM_true_MCS = 10.267260806161483

# Maximum number of seeds to include in plot (x-axis scaling)
seedmaxplot = 205

# Colors and linestyles for the plots. cols assigned to different acquisition 
# functions, cols2 to different indicator quality cases.
linestyles = ['-','--','-.',':']
cols = ['royalblue','indianred','mediumseagreen','orange'] # assigned to the acquisition functions
cols2 = ['goldenrod','maroon','deeppink','deepskyblue'] # assigned to the cases

# Plotswitch for the distributions of all iterations in one plot (1 = plot, 0 = do not plot)
plotalldistroneplot = 1

# Number of iterations to take the mean and limit for max. absolute error E
# (in this case in [m] wave height) in the stopping criterion.
N_mae = 20
lim_MAE_conv = 0.03

# Number of iterations over which to take the coefficient of variation C 
# and associated limit in the stopping criterion.
N_cov = 20
lim_cov_conv = 0.00098

# Acquisition type, including names of acquisition functions in report and last 
# available iteration number. Present numbers were used in the paper. Can be 
# adjusted to match the number of iterations run with 'run_AS' for each combination
# of acquisition function and GPR type.
if GPRvers == 'GPR':
    count_US = 200
    count_USMV = 200
    count_USPD = 200
    count_IVR = 40; count_IVR_BadInd = 100
elif GPRvers == 'MF-GPR':
    count_US = 100
    count_USMV = 200
    count_USPD = 0
    count_IVR = 0


#############
# END INPUT
#############

# Check wether output directories exist, if not make them
isExist1 = os.path.exists(filefolder)
if not isExist1:
   os.makedirs(filefolder)

isExist1 = os.path.exists(figfolder)
if not isExist1:
   os.makedirs(figfolder)


# Target probability of exceedance.
pexp = 1/ne_exp

# Properties attributed to the GPR version and initial sample option
if GPRvers == 'GPR':
    gprcase = 'GPR'
elif GPRvers == 'MF-GPR':
    gprcase = 'MF-GPR'
        
if sampopt == 'A':
    sampopt_txt = 'A_LFsubsamp'; N_hf = 2
elif sampopt == 'B':
    sampopt_txt = 'B_LFfit'; N_hf = 2
elif sampopt == 'C':
    sampopt_txt = 'C_LFisHFfit'; N_hf = 2

# Initialise
plt.figure(10000,figsize=(11,4))
plt.figure(20000,figsize=(5.5,6))

# Loop over the different cases
for nn in np.array((0,1,2,3)):

    # Noise texts (figure titles and names)
    if nn == 0:
        indcase = 'PerfInd'
    elif nn == 1:
        indcase = 'GoodInd'
    elif nn == 2:
        indcase = 'MediumInd'
    elif nn == 3:
        indcase = 'BadInd'
 
    # Load MPM and MPMU95 data
    if GPRvers == 'GPR':
        # header = ['NoIt','MPM_HFGPR','U95_MPM_HFGPR','SumVar_HFGPR']
        Dat_US = pd.read_csv(f'{filefolder}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_US_Nhf{N_hf}_it{count_US}.csv', delimiter=',')
        Dat_USMV = pd.read_csv(f'{filefolder}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_USMV_Nhf{N_hf}_it{count_USMV}.csv', delimiter=',')
        Dat_USPD = pd.read_csv(f'{filefolder}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_USPD_Nhf{N_hf}_it{count_USPD}.csv', delimiter=',')
        # Add IVR data for some
        if nn == 1:
            Dat_IVR = pd.read_csv(f'{filefolder}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_IVR_Nhf{N_hf}_it{count_IVR}.csv', delimiter=',')
        elif nn == 3:
            Dat_IVR = pd.read_csv(f'{filefolder}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_IVR_Nhf{N_hf}_it{count_IVR_BadInd}.csv', delimiter=',')
    elif GPRvers == 'MF-GPR':
        # header = ['NoIt','MPM_HFGPR','U95_MPM_HFGPR','SumVar_HFGPR']
        Dat_US = pd.read_csv(f'{filefolder}/Conv_MF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_US_Nhf{N_hf}_it{count_US}.csv', delimiter=',')
        Dat_USMV = pd.read_csv(f'{filefolder}/Conv_MF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_USMV_Nhf{N_hf}_it{count_USMV}.csv', delimiter=',')

    # Number of samples (not iterations), MPM and U95 of MPM    
    if GPRvers == 'GPR':
        noit_US = Dat_US['NoIt'].to_numpy()
        noit_USMV = Dat_USMV['NoIt'].to_numpy()
        
        MPM_US = Dat_US['MPM_HFGPR'].to_numpy()
        MPM_USMV = Dat_USMV['MPM_HFGPR'].to_numpy() 
                
        MPMU95_US = Dat_US['U95_MPM_HFGPR'].to_numpy()
        MPMU95_USMV = Dat_USMV['U95_MPM_HFGPR'].to_numpy()
    else:
        noit_US = Dat_US['NoIt'].to_numpy()
        noit_USMV = Dat_USMV['NoIt'].to_numpy()
        
        MPM_US = Dat_US['MPM_MFGPR'].to_numpy()
        MPM_USMV = Dat_USMV['MPM_MFGPR'].to_numpy() 
        
        MPMU95_US = Dat_US['U95_MPM_MFGPR'].to_numpy()
        MPMU95_USMV = Dat_USMV['U95_MPM_MFGPR'].to_numpy()
        
        
    # USPD data for all GPR cases 
    if GPRvers == 'GPR':
        noit_USPD = Dat_USPD['NoIt'].to_numpy()
        MPM_USPD = Dat_USPD['MPM_HFGPR'].to_numpy()
        MPMU95_USPD = Dat_USPD['U95_MPM_HFGPR'].to_numpy()
    
        # IVR data for some GPR cases
        if nn == 1 or nn == 3:
            noit_IVR = Dat_IVR['NoIt'].to_numpy()
            MPM_IVR = Dat_IVR['MPM_HFGPR'].to_numpy()
            MPMU95_IVR = Dat_IVR['U95_MPM_HFGPR'].to_numpy()
       
    # Max. number of available iterations
    counts = [count_US,count_USMV,count_USPD,count_IVR]
    itmaxall = max(counts)+1
    
    # Define the colormap Viridis and normalize it to the data range
    cmap = plt.get_cmap('jet') 
    norm = Normalize(vmin=0, vmax=itmaxall)
    
    # Create a ScalarMappable object to map data values to colors
    sm = cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([]) 
    
    # Initialisation of assembled convergence criteria for different acquisition functions
    crit_rej = []
    crit_cov = []
    crit_cov_av = []
    crit_MAE = []
    crit_MAE_av = []
    crit_all = []
    
    # Max. number of acquisition functions to analyse varies per noise case
    # maxnumacq = 6 # not available anymore; IVRMV and IVRPD skipped
    if GPRvers == 'GPR':
        if nn == 1 or nn == 3:
            acqcases = [0,1,2,3] # all 4 acquisition functions
        else:
            acqcases = [0,1,2] # all exclusing IVR
    elif GPRvers == 'MF-GPR':
        acqcases = [0,1] # US and USMV
        # acqcases = [1] # only USMV - messes up the counter for plotting later
    
    
    for mm in acqcases:
        
        # Presently available iterations
        if GPRvers == 'GPR' and mm == 3 and nn == 3:
            itmaxnow = count_IVR_BadInd
        else:
            itmaxnow = counts[mm]
        
        # Initialise assembly of mean predictions
        ass_meanpred = []
        
        # Initialise rejection test for monotonicity
        RejectIt = np.ones(itmaxnow+1)
             
        # initialise MAE 
        D_MAE = np.nan*np.zeros(itmaxnow+1)
        D_MAE_av = np.nan*np.zeros(itmaxnow+1)
        
        # Initialise mean, standard deviation and coefficient of variation of MPM
        stdconv_mpm = np.nan*np.zeros(itmaxnow+1)
        meanconv_mpm = np.nan*np.zeros(itmaxnow+1)
        cov_mpm = np.nan*np.zeros(itmaxnow+1)
        cov_mpm_last = np.nan*np.zeros(itmaxnow+1)
        
        # Initialise final criterion that combines 1,2,3 (starts at 0)
        critall_pa = np.ones(itmaxnow+1)
        
        # Loop over number of iterations
        for ii in range(itmaxnow+1):

            # Color range for increasing it. number
            colnow = [0,0,0.5+(ii/itmaxall)/2]
            
            # Load data - full predicted HF distributions
            if mm == 0:
                if GPRvers == 'GPR':
                    preddata = np.genfromtxt(f'{filefolder}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-US_Nhf{N_hf}_HFit{ii}.csv', skip_header=1, delimiter=',')
                    # Prange_pred = preddata[:,0]
                    # mean_pred = preddata[:,1]
                    # var_pred = GPR1_preddata[:,2]
                    # U95_pred = 1.96*np.sqrt(var_GPR1_pred)
                elif GPRvers == 'MF-GPR':
                    preddata = np.genfromtxt(f'{filefolder}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-US_Nhf{N_hf}_MFit{ii}.csv', skip_header=1, delimiter=',')
                MPMs = MPM_US.copy()
                MPMU95s = MPMU95_US.copy() 
            elif mm == 1:
                if GPRvers == 'GPR':
                    preddata = np.genfromtxt(f'{filefolder}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-USMV_Nhf{N_hf}_HFit{ii}.csv', skip_header=1, delimiter=',')
                elif GPRvers == 'MF-GPR':
                    preddata = np.genfromtxt(f'{filefolder}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-USMV_Nhf{N_hf}_MFit{ii}.csv', skip_header=1, delimiter=',')
                MPMs = MPM_USMV.copy()
                MPMU95s = MPMU95_USMV.copy() 
            elif mm == 2:
                if GPRvers == 'GPR':
                    preddata = np.genfromtxt(f'{filefolder}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-USPD_Nhf{N_hf}_HFit{ii}.csv', skip_header=1, delimiter=',')
                MPMs = MPM_USPD.copy()
                MPMU95s = MPMU95_USPD.copy() 
            elif mm == 3:
                if GPRvers == 'GPR':
                    preddata = np.genfromtxt(f'{filefolder}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-IVR_Nhf{N_hf}_HFit{ii}.csv', skip_header=1, delimiter=',')
                MPMs = MPM_IVR.copy()
                MPMU95s = MPMU95_IVR.copy() 
                    
            # Extract PoE range and mean predictions (HF version MF-GPR at same columns as GPR predictions)
            Prange_pred = preddata[:,0]
            mean_pred = preddata[:,1]
            
            # MPM and MPM U95 values up to the present iteration
            mpms_now = MPMs[0:ii]
            u95mpms_now = MPMU95s[0:ii]
            
            ##################  
            # Crit1: Reject non-monotonic results 
            ################## 
              
            # Check whether prediction in present iteration is monotonically decreasing, 
            # and that there is a difference between top and down of more than 1% of top value
            eps1 = 0.01*mean_pred[0]
            if all(mean_pred[i] > mean_pred[i+1] for i in range(len(mean_pred)-1)):
                # Yes, monotonically decreasing
                if mean_pred[0]-mean_pred[-1] > eps1:
                    # Mon. decreasing, but difference top-down big enough
                    RejectIt[ii] = 0
                else:
                    # Mon. decreasing, but difference top-down too small
                    RejectIt[ii] = 1
            else:
                # No, not monotonically decreasing
                RejectIt[ii] = 1
            
            ################## End crit1

            # Plot all distributions on top of each other
            if plotalldistroneplot == 1:
                plt.figure(2000+100*mm+10*nn)
                if RejectIt[ii] == 1:
                    plt.plot(mean_pred,Prange_pred,color='grey')
                    plt.text(mean_pred[-1],Prange_pred[-1],f'{ii}',color='grey')                    
                else:
                    if np.mod(ii,20) == 0:
                        plt.plot(mean_pred,Prange_pred,color=sm.to_rgba(ii),label=f'Iteration {ii}')
                    else:
                        plt.plot(mean_pred,Prange_pred,color=sm.to_rgba(ii))
                    plt.text(mean_pred[-1],Prange_pred[-1],f'{ii}',color=sm.to_rgba(ii))
                        
            # Save these data to use in convergence assessment
            ass_meanpred.append(mean_pred)
            
            ##################  
            # Crit2 alt: Calculate maximum absolute error
            ##################     
                                
            # Loop starts from 2 available iterations.
            if ii > 0:
                
                # DNR present and previous iteration
                pred1 = ass_meanpred[ii-1]
                pred2 = ass_meanpred[ii]
                
                # max absolute difference
                # MAE = np.mean(np.abs(pred1-pred2))
                MAE = np.max(np.abs(pred1-pred2))
                # MAE = np.sqrt(np.mean((pred1-pred2)**2))
                
                # Assemble data
                D_MAE[ii] = MAE
                        
            # Calculate the MAE over the last N_mae iterations
            # Start at > 1 because D_MAE[:0] = [] and D_MAE[:1] = D_MAE[0] = nan
            if ii > 1:
                if ii <= N_mae-1:
                    # "Burn-in": discard portion of simulation results to ensure steady state (or treat them differently)
                    D_mae_t = D_MAE[:ii][np.isfinite(D_MAE[:ii])]
                elif ii > N_mae-1:
                    D_mae_t = D_MAE[ii-N_mae:ii][np.isfinite(D_MAE[ii-N_mae:ii])]
                D_MAE_av[ii] = sum(D_mae_t)/len(D_mae_t)
              
           
           ################## End crit2 alt
           
            ##################  
            # Crit3: Calculate standard deviation and COV of MPM over existing iterations
            ##################     
              
            # Sample standard deviation
            stdconv_mpm[ii] = np.std(mpms_now)
            
            # Sample mean
            meanconv_mpm[ii] = np.mean(mpms_now)  
            
            # Coefficient of variation up to present iteration
            cov_mpm[ii] = stdconv_mpm[ii]/meanconv_mpm[ii]

            # Very erratic; calculate the COV over the last N_cov iterations
            # Start at > 0 because mpms_now[:0] = []
            if ii > 1:
                # N_cov = 10
                if ii <= N_cov-1:
                    # "Burn-in": discard portion of simulation results to ensure steady state (or treat them differently)
                    mpms_t = mpms_now[:ii][np.isfinite(mpms_now[:ii])]
                elif ii > N_cov-1:
                    mpms_t = mpms_now[ii-N_cov:ii][np.isfinite(mpms_now[ii-N_cov:ii])]
                # cov_mpm_last[ii] = (np.std(mpms_t)/np.sqrt(len(mpms_t)))/np.mean(mpms_t)  
                cov_mpm_last[ii] = np.std(mpms_t)/np.mean(mpms_t)   
                # cov_mpm_last[ii] = np.std(mpms_t)
            
            ################## End crit3
                       
            
            
            ###########
            # One criterion that combines them all
            ###########
                
            # First check monotonicity criterion
            if RejectIt[ii] == 1:
                critall_pa[ii] = 0
           
            # Check the mean absolute error convergence criterion
            if D_MAE_av[ii] > lim_MAE_conv:
                critall_pa[ii] = 0
            elif np.isnan(D_MAE_av[ii]):
                critall_pa[ii] = 0  
                
            # Check the MPM / quantile COV convergence criterion
            if cov_mpm_last[ii] > lim_cov_conv:
                critall_pa[ii] = 0
            elif np.isnan(cov_mpm_last[ii]):
                critall_pa[ii] = 0
          

        # Finish figure of all distributions
        if plotalldistroneplot == 1:
            plt.figure(2000+100*mm+10*nn)
            plt.yscale('log')
            plt.title(f'Indicator quality case: {indcase} \n {gprcase} (grey = rejected iteration)')
            plt.ylabel('P(c$\geq$C) [-]')
            plt.xlabel('C [m]')
            plt.ylim([4*10**(-4),2*10**(-1)]) # room for legend
            plt.xlim([7.5,12.5])
            plt.legend(ncol=2)
            plt.plot([0,15],[pexp,pexp],':',color='black')
            plt.text(12.1,pexp*1.2,'$P_{exp}$')
            plt.savefig(f'{figfolder}/AllDistrForIts_{indcase}_{gprcase}_VarAcq.pdf',bbox_inches='tight')
        
        # Save for all acquisition functions
        crit_rej.append(RejectIt)
        crit_cov.append(cov_mpm)
        crit_cov_av.append(cov_mpm_last)
        crit_MAE.append(D_MAE)
        crit_MAE_av.append(D_MAE_av)
        crit_all.append(critall_pa)
    
     
    ###########
    # First converged iteration number
    ###########
    
    ic_US = np.argmax(crit_all[0]>0)
    ic_USMV = np.argmax(crit_all[1]>0)
    if GPRvers == 'GPR': 
        ic_USPD = np.argmax(crit_all[2]>0)
        if nn == 1 or nn == 3:
            ic_IVR = np.argmax(crit_all[3]>0)
    
    
    ###########
    # Some plot settings
    ###########
    
    # Location of text 'Criterion'
    x_loc_crit = 170
    x_ticks = np.arange(0,202,25)
    
    # Set the rejected iterations to NaN in the other criteria plots
    nanmask_rej_US = np.where(crit_rej[0]==1,np.nan,1)
    nanmask_rej_USMV = np.where(crit_rej[1]==1,np.nan,1)
    if GPRvers == 'GPR':
        nanmask_rej_USPD = np.where(crit_rej[2]==1,np.nan,1)
        if nn == 1 or nn == 3:
            nanmask_rej_IVR = np.where(crit_rej[3]==1,np.nan,1)
    
    ###########
    # MPM plot for all available iterations
    ###########
    
   
    # Same but all in one plot
    plt.figure(10000)
    plt.subplot(2,2,nn+1)
    plt.plot(noit_US,MPM_US*nanmask_rej_US,linestyle=linestyles[0],color=cols[0],label='US')  
    plt.plot(noit_USMV,MPM_USMV*nanmask_rej_USMV,linestyle=linestyles[1],color=cols[1],label='USMV')
    if GPRvers == 'GPR':
        plt.plot(noit_USPD,MPM_USPD*nanmask_rej_USPD,linestyle=linestyles[2],color=cols[2],label='USPD')      
        if nn == 1 or nn == 3:
            plt.plot(noit_IVR,MPM_IVR*nanmask_rej_IVR,linestyle=linestyles[3],color=cols[3],label='IVR')
        else:
            plt.plot(0,0,linestyle=linestyles[3],color=cols[3],label='IVR')
    else:
        plt.plot(0,0,linestyle=linestyles[2],color=cols[2],label='USPD')      
        plt.plot(0,0,linestyle=linestyles[3],color=cols[3],label='IVR')
    plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey',linewidth=0.8) # ,label='Truth (MCS)')
    if nn == 2 or nn == 3:
        plt.text(165,10.30,'MCS \'truth\'',color='grey')
        plt.xticks(ticks=x_ticks)
        plt.xlabel('Number of samples $j$ [-]')
    else:
        plt.text(165,10.1,'MCS \'truth\'',color='grey')
        plt.xticks(ticks=x_ticks,labels=[])
    plt.xlim([0,seedmaxplot])
    plt.yticks(ticks=[9,9.5,10,10.5,11,11.5])
    plt.ylim([9.5,11])
    if nn == 0:
        plt.legend(ncol=4)
    if nn == 0 or nn == 2:
        plt.ylabel('$\widehat{C\'\'}$ [m]')
    plt.title(indcase)

    # ###########
    # # Final plot including convergence criteria
    # ###########

    # Details for the selected acquisition function (only implemented if this 
    # is US or USMV, but others can easily be added)
    if acqsel == 'US':
        noit_sel = noit_US
        MPM_sel = MPM_US
        nanmask_rej_sel = nanmask_rej_US
        crit_MAE_av_sel = crit_MAE_av[0]
        crit_cov_av_sel = crit_cov_av[0]
        ic_sel = ic_US
    elif acqsel == 'USMV':
        noit_sel = noit_USMV
        MPM_sel = MPM_USMV
        nanmask_rej_sel = nanmask_rej_USMV
        crit_MAE_av_sel = crit_MAE_av[1]
        crit_cov_av_sel = crit_cov_av[1]
        ic_sel = ic_USMV
        
    # Convergence criterion plot 
    plt.figure(20000)
    
    # MPM    
    ax1 = plt.subplot(3,1,1)
    plt.ylabel('$\widehat{C\'\'}$ [m]')
    plt.plot(noit_sel[:ic_sel-1],MPM_sel[:ic_sel-1],linestyle=linestyles[nn],color=cols2[nn],label=indcase)  
    plt.plot(noit_sel[ic_sel:],MPM_sel[ic_sel:],linestyle=linestyles[nn],color=cols2[nn],alpha=0.4)  
    plt.plot(noit_sel[ic_sel],MPM_sel[ic_sel],linestyle='none',color=cols2[nn],marker = 'x') 
    if nn == 3:
        plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='grey',linewidth=0.8,label='MCS \'truth\'')
        plt.yticks(ticks=[9,9.5,10,10.5,11,11.5])
        plt.ylim([9,11])
        plt.xlim([0,seedmaxplot])
        plt.xticks(ticks=x_ticks,labels=[])
        if GPRvers == 'GPR':
            plt.legend(ncol=3,loc='lower right')
    
    # MAE criterion 2alt
    ax2 = plt.subplot(3,1,2)
    plt.ylabel(r'$\overline{E}_{%i}$ [m]' % (N_mae))
    plt.plot(noit_sel,crit_MAE_av_sel*nanmask_rej_sel,linestyle=linestyles[nn],color=cols2[nn],label=indcase)
    if nn == 3:
        plt.xlim([0,seedmaxplot])
        plt.plot([0,seedmaxplot],[lim_MAE_conv,lim_MAE_conv],linestyle='--',color='grey',linewidth=0.8)
        plt.text(x_loc_crit,lim_MAE_conv*1.5,'Criterion',color='grey')
        plt.yscale('log')
        plt.ylim([10**(-4),1])
        plt.xticks(ticks=x_ticks,labels=[])
           
    # MPM COV criterion 3 (plus averaged over a number of iterations)
    ax3 = plt.subplot(3,1,3)
    plt.ylabel(r'$C_{%i}$ [-]' % (N_cov))
    plt.plot(noit_sel,crit_cov_av_sel*nanmask_rej_sel,linestyle=linestyles[nn],color=cols2[nn],label=indcase)
    if nn == 3:
        plt.xlim([0,seedmaxplot])
        plt.plot([0,seedmaxplot],[lim_cov_conv,lim_cov_conv],linestyle='--',color='grey',linewidth=0.8)
        plt.text(x_loc_crit,lim_cov_conv*1.5,'Criterion',color='grey')
        plt.yscale('log')
        plt.ylim([10**(-5),10**(-1)])
        plt.xticks(ticks=x_ticks)
        plt.xlabel('Number of samples $j$ [-]')

    ###### Print results to screen
    print(f'FULL CONVERGENCE WITH PRESENT CRITERIA - {gprcase}:')
    print(f'Case {indcase}, US: converged at {ic_US} iterations - {ic_US+N_hf} samples (of max. {count_US})')
    print(f'Case {indcase}, USMV: converged at {ic_USMV} iterations - {ic_USMV+N_hf} samples (of max. {count_USMV})')
    if GPRvers == 'GPR':
        print(f'Case {indcase}, USPD: converged at {ic_USPD} iterations - {ic_USPD+N_hf} samples (of max. {count_USPD})')
        if nn == 1:
            print(f'Case {indcase}, IVR: converged at {ic_IVR} iterations - {ic_IVR+N_hf} samples (of max. {count_IVR})')
        elif nn == 3:
            print(f'Case {indcase}, IVR: converged at {ic_IVR} iterations - {ic_IVR+N_hf} samples (of max. {count_IVR_BadInd})')
    
# Finalise and save plot
plt.figure(10000)
plt.savefig(f'{figfolder}/Fig10_Vary_acquisition_AllAcq_{gprcase}_Set-{HFkern}-nf_{sampopt_txt}_Nhf{N_hf}_MPM.pdf',bbox_inches='tight')

plt.figure(20000)
plt.savefig(f'{figfolder}/Fig11_ConvergencePlots_{gprcase}_acq{acqsel}_Set-{HFkern}-nf_{sampopt_txt}_Nhf{N_hf}.pdf',bbox_inches='tight')

   