# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen

"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import glob
import os
import re

# Add some other functions
from functions.subsample_routines import subsample_binned
from functions.distribution_scripts import MPM_from_DNR
from functions.theoretical_distributions import make_cdf_Forristall as Forr

# Convergence criterion
from functions.convergenceAS import stopping_crit_AS

# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Plotsize
fsize_h = 7
fsize_v = 5

# Number of seeds to include in plot
seedmaxplot = 201
seedstep = 20

# Acquisition function type to evaluate.
acqnames = ['USMV']

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'B'

# Folder name where the output of the Adaptive Screening procedure is saved.
# In the default case this is the same folder as where we write the output, but
# this can be defined differently if preferred.
in_files = 'out_files'

# Folder names in which the CSV files generated by Adaptive Screening were stored, 
# and name of the folder in which new figures will be stored. Defined relative 
# to the present folder, and the folder with this name needs to exist.
out_figs = 'out_figs'

# Corresponding line styles and colors in the plots
stylesgpr = ['-.'];     colsgpr = ['green']  # matching distribution plots
stylesmfgpr = ['--'];   colsmfgpr = ['blue'] # matching distribution plots

# Kernel name single-fidelity HF-GPR (preferred Mat32 for all)
HFkern = 'Mat32'
MFkern1 = 'Mat32' 
MFkern2 = 'Mat32'

# Spectral properties basis wave system (for ref. Forristall and Pexp)
dur = 3600 # target duration
Hs = 10 # sign wave height
Tp = 11 # peak wave period
wd = 30 # water depth
Vs = 0 # current speed or vessel speed
T1 = Tp/1.198 # estimated from relation for JONSWAP

# Forristall
VrangeRF = np.linspace(0, 15, 101)[:, None]

# Annotation switch for the plot of the MPM as function of iterations until 
# convergence. Paper figure 12 includes annotation.
# 1 = with annotation, 0 = without annotation
annon = 1

# Switch for plotting of the intermediate results as in appendix
# 1 = plots of intermediate iterations, 0 = no plots of intermediate iterations
plotintermediateresults = 0

#############
# END INPUT
#############


# Check wether output directory exist, if not make it
isExist1 = os.path.exists(out_figs)
if not isExist1:
   os.makedirs(out_figs)

# Initialise
plt.figure(10000,figsize=(12.5,6))
x_ticks = np.arange(0,202,25)

# Initial sample options (see Adaptive Screening)
if sampopt == 'A':
    sampopt_txt = 'A_LFsubsamp'; N_hf = 2
elif sampopt == 'B':
    sampopt_txt = 'B_LFfit'; N_hf = 2
elif sampopt == 'C':
    sampopt_txt = 'C_LFisHFfit'; N_hf = 2

# Derive target probability of exceedance, using the average number of wave encounters in target duration
ne_exp = dur/Tp 
pexp = 1/ne_exp

# Calc Forristall ref distribution
p_For,dum2 = Forr(VrangeRF,Hs,T1,wd,Vs)

# Initialise
conv_mpm_gpr = np.nan*np.zeros((4,len(acqnames)))
conv_mpmu95_gpr = np.nan*np.zeros((4,len(acqnames)))
conv_mpm_mfgpr = np.nan*np.zeros((4,len(acqnames)))
conv_mpmu95_mfgpr = np.nan*np.zeros((4,len(acqnames)))

# Loop over conditions
for nn in range(4):
# for nn in np.array([1,2,3]):
    
    # Noise texts (figure titles and names)
    if nn == 0:
        noisetxt = 'PerfInd'
    elif nn == 1:
        noisetxt = 'GoodInd'
    elif nn == 2:
        noisetxt = 'MediumInd'
    elif nn == 3:
        noisetxt = 'BadInd'

    # Load true data (reference for DNR plots, and to derive true MPM)
    DatLFtrue = np.loadtxt(f'{in_files}/Lin2nd_trueLFdata_{noisetxt}_{sampopt_txt}.csv', delimiter=',')
    v_true_l = DatLFtrue[:,0]
    p_true_l = DatLFtrue[:,1]
    DatHFtrue = np.loadtxt(f'{in_files}/Lin2nd_trueHFdata_{noisetxt}_{sampopt_txt}.csv', delimiter=',')
    v_true_h = DatHFtrue[:,0]
    p_true_h = DatHFtrue[:,1]
    DatLFtrain = np.loadtxt(f'{in_files}/Lin2nd_trainLFdata_{noisetxt}_{sampopt_txt}.csv', delimiter=',')
    v_train_l = DatLFtrain[:,0]
    p_train_l = DatLFtrain[:,1]
    DatHFtraininit = np.loadtxt(f'{in_files}/Lin2nd_trainHFdata_init_{noisetxt}_{sampopt_txt}.csv', delimiter=',')
    v_train_h_init = DatHFtraininit[:,0]
    p_train_h_init = DatHFtraininit[:,1]
    
    # Subsample the MCS true distribution data for plotting later
    # binsize = 0.05
    binsize = 0.02
    bins = np.arange(0,15,binsize)
    v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,bins)
    v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,bins)
    
    # Calculate / derive true MPM
    MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp)
    if noisetxt == 'NoNoise':
        MPM_true_For = MPM_from_DNR(VrangeRF,p_For,ne_exp)
    else:
        MPM_true_For = np.nan
    
    # Loop over selected acquisition functions (only functions that are available for both GPR and MF-GPR)
    for mm in range(len(acqnames)):
        acqname = acqnames[mm]
        print(f'Case {noisetxt} ({nn}/4), acq. function {acqname} ({mm}/{len(acqnames)-1})')
    
        #### Find last available iteration (in this case probably #200, as we later interpolate in this database using a convergence criterion)
    
        # List all files in directory with the proper name, then select last available iteration
        # Getting the list of files/directories in the specified path, filtering the list to exclude the directory names 
        files_hf = list(filter(os.path.isfile, glob.glob(f'{in_files}/Conv_HF-GPR_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it*.csv')))
        files_mf = list(filter(os.path.isfile, glob.glob(f'{in_files}/Conv_MF-GPR_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it*.csv')))
      
        # Find last available number of iterations
        ass_count_hf = np.array([],dtype=int)
        for file_i in files_hf:
            temp = int(re.search(r'\d+', file_i[-8:-1]).group())
            ass_count_hf = np.append(ass_count_hf,temp)
        count_hf_all = np.max(ass_count_hf)   
        del(temp)
    
        ass_count_mf = np.array([],dtype=int)
        for file_i in files_mf:
            temp = int(re.search(r'\d+', file_i[-8:-1]).group())
            ass_count_mf = np.append(ass_count_mf,temp)
        count_mf_all = np.max(ass_count_mf)    
        del(temp)
      
        print(f'{noisetxt} with HF-GPR and {acqname}: available are {count_hf_all} iterations ({count_hf_all+N_hf} samples)')
        print(f'{noisetxt} with MF-GPR and {acqname}: available are {count_mf_all} iterations ({count_mf_all+N_hf} samples)')
        
        
        #### Load data for this last iteration and plot convergence
    
        # Load data HF-GPR
        Dat_HF_GPR_all = pd.read_csv(f'{in_files}/Conv_HF-GPR_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it{count_hf_all}.csv', delimiter=',')
        Dat_MF_MFGPR_all = pd.read_csv(f'{in_files}/Conv_MF-GPR_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it{count_mf_all}.csv', delimiter=',') # foutje in kernel names (Hf ipv 2xMF), maar al in andere file gemaakt.
       
        # Number of samples! (not iterations)
        noit_GPR = Dat_HF_GPR_all['NoIt'].to_numpy()
        noit_MFGPR = Dat_MF_MFGPR_all['NoIt'].to_numpy()
    
        # Actual MPM values
        mpms_GPR = Dat_HF_GPR_all['MPM_HFGPR'].to_numpy()
        mpmsU95_GPR = Dat_HF_GPR_all['U95_MPM_HFGPR'].to_numpy()
        
        mpms_MFGPR = Dat_MF_MFGPR_all['MPM_MFGPR'].to_numpy()
        mpmsU95_MFGPR = Dat_MF_MFGPR_all['U95_MPM_MFGPR'].to_numpy()
               
        # # Plot convergence behaviour for all available iterations (can be 
        # useful to uncomment if convergence seems not proper for a new case - 
        # in that case also uncomment the lines below 'Save the convergence plots 
        # per case' below). 
        # fig100 = plt.figure(100+10*nn,figsize=(6.1,3))
        # plt.ylabel('$\widehat{C\'\'}$ [m]')
        # plt.xlabel('Number of samples $j$ [-]')
        # plt.plot(noit_GPR,mpms_GPR,linestyle=stylesgpr[mm],color=colsgpr[mm],label=f'GPR')  
        # plt.plot(noit_MFGPR,mpms_MFGPR,linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label=f'MF-GPR')    
        # plt.fill_between(noit_GPR.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
        #                   (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsgpr[0], alpha=0.3,label=f'GPR U95%')
        # plt.fill_between(noit_MFGPR.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
        #                   (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label=f'MF-GPR U95%')
        # plt.xticks(np.arange(0,seedmaxplot,seedstep,dtype=int))
        # plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='black')
        # plt.text(seedmaxplot-30,10.08,'True MCS',color='black')
        # plt.ylim([9,12])
        # plt.xlim([N_hf-1,seedmaxplot])
        # # plt.title(f'Indicator quality case: {noisetxt2}')
        # plt.xticks(ticks=np.arange(0,seedmaxplot,seedstep,dtype=int),label=[])
        # plt.legend(ncol=2)
        
        #### Apply stopping criterion to determine at which iteration we are converged
        
        # File name start of CSV with predicted data per iteration (exclude "itXX.csv", will be added later)
        fn_preds_GPR = f'{in_files}/HFGPR_HFpred_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-{acqname}_Nhf{N_hf}_HF'
        casename_GPR = f'{noisetxt} - GPR - {acqname}'
        min_it1 = 5
        ploton1 = False
        convit1 = stopping_crit_AS(noit_GPR,mpms_GPR,mpmsU95_GPR,fn_preds_GPR,casename_GPR,pexp,min_it1,ploton1) # final
        
        fn_preds_MFGPR = f'{in_files}/MFGPR_HFpred_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MF'
        casename_MFGPR = f'{noisetxt} - MF-GPR - {acqname}'
        min_it2 = 5
        ploton2 = False
        convit2 = stopping_crit_AS(noit_MFGPR,mpms_MFGPR,mpmsU95_MFGPR,fn_preds_MFGPR,casename_MFGPR,pexp,min_it2,ploton2) # final

        # Prepare for annotation
        if annon == 1:
            gpr10 = np.round((mpms_GPR[9]/MPM_true_MCS-1)*100,1);  signgpr10 = ''
            if gpr10>0: 
                signgpr10 = '+' 
            mfgpr10 = np.round((mpms_MFGPR[9]/MPM_true_MCS-1)*100,1); signmfgpr10 = ''
            if mfgpr10>0: 
                signmfgpr10 = '+' 
            gprconv = np.round((mpms_GPR[convit1]/MPM_true_MCS-1)*100,1); signgprconv = ''
            if gprconv>0: 
                signgprconv = '+' 
            mfgprconv = np.round((mpms_MFGPR[convit2]/MPM_true_MCS-1)*100,1); signmfgprconv = ''
            if mfgprconv>0: 
                signmfgprconv = '+'  
            
        # Same but all in one plot
        plt.figure(10000)
        plt.subplot(2,2,nn+1) 
        plt.title(noisetxt)
        if nn == 0 or nn == 2:
            plt.ylabel('$\widehat{C\'\'}$ [m]')
        if nn == 2 or nn == 3:
            plt.xlabel('Number of samples $j$ [-]')
            plt.xticks(ticks=x_ticks)
        else:
            plt.xticks(ticks=x_ticks,labels=[])
        plt.plot([0,seedmaxplot],0.98*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3); plt.text(seedmaxplot-12,0.98*MPM_true_MCS+0.03,'-2%',color='lightgrey',fontsize=8)
        plt.plot([0,seedmaxplot],1.02*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3); plt.text(seedmaxplot-12,1.02*MPM_true_MCS+0.03,'+2%',color='lightgrey',fontsize=8)
        plt.plot([0,seedmaxplot],0.96*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3); plt.text(seedmaxplot-12,0.96*MPM_true_MCS+0.03,'-4%',color='lightgrey',fontsize=8)
        plt.plot([0,seedmaxplot],1.04*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3); plt.text(seedmaxplot-12,1.04*MPM_true_MCS+0.03,'+4%',color='lightgrey',fontsize=8)
        plt.plot(noit_GPR[:convit1],mpms_GPR[:convit1],linestyle=stylesgpr[mm],color=colsgpr[mm],label='GPR')  
        plt.plot(noit_MFGPR[:convit2],mpms_MFGPR[:convit2],linestyle=stylesmfgpr[mm],color=colsmfgpr[mm],label='MF-GPR')    
        plt.fill_between(noit_GPR[:convit1].flatten(), (mpms_GPR[:convit1] - 0.5*mpmsU95_GPR[:convit1]).flatten(), 
                          (mpms_GPR[:convit1] + 0.5*mpmsU95_GPR[:convit1]).flatten(), facecolor=colsgpr[0], alpha=0.3,label='GPR U95%')
        plt.fill_between(noit_MFGPR[:convit2].flatten(), (mpms_MFGPR[:convit2] - 0.5*mpmsU95_MFGPR[:convit2]).flatten(), 
                          (mpms_MFGPR[:convit2] + 0.5*mpmsU95_MFGPR[:convit2]).flatten(), facecolor=colsmfgpr[0], alpha=0.3,label='MF-GPR U95%')
        plt.xticks(np.arange(0,seedmaxplot,seedstep,dtype=int))
        plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],linestyle=':',color='black'); plt.text(seedmaxplot-30,MPM_true_MCS+0.03,'True MCS',color='black')
        plt.ylim([9,12])
        plt.xlim([N_hf-1,seedmaxplot])
        plt.xticks(ticks=np.arange(0,seedmaxplot,seedstep,dtype=int),label=[])
        if nn == 0:
            plt.legend(ncol=2)
        if annon == 1:
            plt.plot([noit_GPR[8],noit_GPR[8]],[MPM_true_MCS,11.8],color='grey',linewidth=0.2)
            plt.text(noit_GPR[8]+1,11.3,f'10 samples\nGPR: {signgpr10}{gpr10}%\nMF-GPR: {signmfgpr10}{mfgpr10}%',color='grey',fontsize=8)
            plt.plot([noit_GPR[convit1]-1,noit_GPR[convit1]-1],[mpms_GPR[convit1],11.05],color='green',linewidth=0.2)
            plt.text(noit_GPR[convit1],10.9,f'converged\nGPR: {signgprconv}{gprconv}%',color='green',fontsize=8)
            plt.plot([noit_MFGPR[convit2]-1,noit_MFGPR[convit2]-1],[mpms_MFGPR[convit2],9.25],color='blue',linewidth=0.2)
            plt.text(noit_MFGPR[convit2],9.2,f'converged\nMF-GPR: {signmfgprconv}{mfgprconv}%',color='blue',fontsize=8)
            
        #### CALCULATE AND PRINT MPM VALUES AND U95% FOR CONVERGED RESULTS
        conv_mpm_gpr[nn,mm] = mpms_GPR[convit1]
        conv_mpmu95_gpr[nn,mm] = mpmsU95_GPR[convit1]
        conv_mpm_mfgpr[nn,mm] = mpms_MFGPR[convit2]
        conv_mpmu95_mfgpr[nn,mm] = mpmsU95_MFGPR[convit2]
        
        print(f'Case: {noisetxt}, {dur/3600}h:')
        print(f'MPM True MCS: {MPM_true_MCS} m')
        print(f'MPM True Forristall: {MPM_true_For} m')
        print(f'MPM GPR {acqname}: {conv_mpm_gpr[nn,mm]} m, U95 {conv_mpmu95_gpr[nn,mm]} m')
        print(f'MPM MF-GPR {acqname}: {conv_mpm_mfgpr[nn,mm]} m, U95 {conv_mpmu95_mfgpr[nn,mm]} m')

        ############### READ AND PLOT CONVERGED GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        
        # GPR (2x)
        DatHFtrainIt1 = np.genfromtxt(f'{in_files}/HFGPR_TrainData_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-{acqname}_Nhf{N_hf}_HFit{convit1}.csv', skip_header=1, delimiter=',')
        v_train_h1 = DatHFtrainIt1[:,0]
        p_train_h1 = DatHFtrainIt1[:,1]
        
        DatHFpredIt1 = np.genfromtxt(f'{in_files}/HFGPR_HFpred_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-{acqname}_Nhf{N_hf}_HFit{convit1}.csv', skip_header=1, delimiter=',')
        Prange_predHF = DatHFpredIt1[:,0]
        mean_high_gp_model = DatHFpredIt1[:,1]
        var_high_gp_model = DatHFpredIt1[:,2]
        U95_high_gp_model = 1.96*np.sqrt(var_high_gp_model)
        
        ###### Plot HF
        # Plot the posterior mean and variance (U95) from the reference HF gaussian process
        fig10 = plt.figure(1000,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS')
        plt.plot(v_train_h1,p_train_h1, color='r', marker='d',linestyle='None', label=f'{convit1} HF samples')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                          (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction U95%')
        plt.plot([0,15],[pexp,pexp],':',color='black')
        plt.text(14,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.ylim([10**(-5),1])
        plt.xlim([0,15])
        plt.legend()
        plt.savefig(f'{out_figs}/Fig13_Prediction_HFGPR_{noisetxt}_Set-{HFkern}_{sampopt_txt}_{acqname}_it{convit1}.pdf',bbox_inches='tight')
       
        
        
        ############### READ AND PLOT CONVERGED MF-GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        
        # MF-GPR (4x)
        DatLFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFTrainData_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        v_train_l2 = DatLFtrainIt2[:,0]
        p_train_l2 = DatLFtrainIt2[:,1]
        
        DatHFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFTrainData_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        v_train_h2 = DatHFtrainIt2[:,0]
        p_train_h2 = DatHFtrainIt2[:,1]
        
        DatLFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFpred_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        Prange_predLF = DatLFpredIt2[:,0]
        mean_low_mfgp_model = DatLFpredIt2[:,1]
        var_low_mfgp_model = DatLFpredIt2[:,2]
        U95_low_mfgp_model = 1.96*np.sqrt(var_low_mfgp_model)
        
        DatHFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFpred_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
        Prange_predHF = DatHFpredIt2[:,0]
        mean_high_mfgp_model = DatHFpredIt2[:,1]
        var_high_mfgp_model = DatHFpredIt2[:,2]
        U95_high_mfgp_model = 1.96*np.sqrt(var_high_mfgp_model)
        
        ###### Plot MF
        # Plot the posterior mean and variance (U95) from the MF gaussian process
        fig20 = plt.figure(2000,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS')
        plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
        plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None', label=f'{convit2} HF samples')
        plt.plot(mean_low_mfgp_model,Prange_predLF,'--', color='b',label='MF-GPR predictions')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF (label not used)')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                          (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
        plt.plot([0,15],[pexp,pexp],':',color='black')
        plt.text(14,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.ylim([10**(-5),1])
        plt.xlim([0,15])
        handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
        order = [0,1,2,3,4,5] #specify order of items in legend
        plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
        plt.savefig(f'{out_figs}/Fig14_Prediction_MFGPR_{noisetxt}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_it{convit2}.pdf',bbox_inches='tight')
            
        
       
        
        ############### PLOT GPR AND MF-GPR IN SAME PLOT
        
        # ALL IN ONE Plot the posterior mean and variance (U95) from the MF gaussian process (without training data, as these are different for HF-GPR and MF-GPR)
        fig30 = plt.figure(30,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS') 
        plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',label='MF-GPR predictions')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                          (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF  (label not printed)')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                          (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                          (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction HF U95%')
        plt.plot([0,15],[pexp,pexp],':',color='black')
        plt.text(14,pexp*1.4,'$P_{exp}$')
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.ylim([10**(-5),1])
        plt.xlim([0,15])
        # plt.title(f'Adaptive GRR and MF-GPR \n GPR converged at it.{convit1} ({convit1+N_hf} samples) and MF-GPR at it.{convit2} ({convit2+N_hf} samples)')
        handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
        order = [0,1,2,3,6,7] #specify order of items in legend
        plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
        plt.savefig(f'{out_figs}/AllPred_{noisetxt}_HFSet-{HFkern}_MFSet-{MFkern1}-{MFkern2}_HF-{sampopt_txt}_MF-{sampopt_txt}_HFacq-{acqname}_MFacq-{acqname}_HFit{convit1}_MFit{convit2}.pdf',bbox_inches='tight')

        plt.close(fig10); plt.close(fig20); plt.close(fig30)
        
        
        # ####################################
        # # Calculate maximum absolute difference between predicted and true HF distributions
        # ####################################
        # Not used in the paper, was considered as additional quality metric.
        
        # # The mean predictions mean_high_gp_model (GPR) and mean_high_mfgp_model (MF-GPR) 
        # # are both defined on probability axis Prange_predHF. The true MCS result 
        # # v_true_h is defined on another probability axis, p_true_h. Use linear 
        # # interpolation of the true values on the predicted axis to compare them directly.
        
        # # One-dimensional linear interpolation for monotonically increasing sample 
        # # points (so first flip the probability axis). y = numpy.interp(x, xp, fp)
        # v_true_h_interp = np.interp(Prange_predHF,np.flip(p_true_h),np.flip(v_true_h))
        
        # # # Check1: interpolation of truth
        # # plt.figure()
        # # plt.yscale('log')
        # # plt.plot(v_true_h,p_true_h,'g')
        # # plt.plot(v_true_h_interp,Prange_predHF,'xg')
        
        # # # Check2: interpolated truth and predictions
        # # plt.figure()
        # # plt.yscale('log')
        # # plt.plot(v_true_h_interp,Prange_predHF,'xg')
        # # plt.plot(mean_high_gp_model,Prange_predHF,'+b')
        # # plt.plot(mean_high_mfgp_model,Prange_predHF,'*r')
        
        # # Calculate the max. absolute difference values
        # MeaAD_gpr = np.mean(np.abs(mean_high_gp_model-v_true_h_interp))
        # MaxAD_gpr = np.max(np.abs(mean_high_gp_model-v_true_h_interp))
        # MeaAD_mfgpr = np.mean(np.abs(mean_high_mfgp_model-v_true_h_interp))
        # MaxAD_mfgpr = np.max(np.abs(mean_high_mfgp_model-v_true_h_interp))
          
        # print(f'Max. abs. diff. GPR {acqname} from truth: {MaxAD_gpr} m')
        # print(f'Max. abs. diff. MF-GPR {acqname} from truth: {MaxAD_mfgpr} m')
        
        
    # # Save the convergence plots per case
    # plt.figure(100+10*nn)
    # plt.savefig(f'{out_figs}/Convergence_{noisetxt}_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_allit.pdf',bbox_inches='tight')           

    ####################################
    # Also save a few intermediate plots
    ####################################
    
    if plotintermediateresults == 1:
    
        # This is already done in version 'Plot_results_ConvAndDNR_interpolate8' > the fully converged version.
        intermedind = [0,5,10,20,50,100,150,200]
        
        for bb in intermedind:
                
            ############### READ AND PLOT INTERMEDIATE GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
            
            # GPR (2x)
            DatHFtrainIt1 = np.genfromtxt(f'{in_files}/HFGPR_TrainData_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-{acqname}_Nhf{N_hf}_HFit{bb}.csv', skip_header=1, delimiter=',')
            v_train_h1 = DatHFtrainIt1[:,0]
            p_train_h1 = DatHFtrainIt1[:,1]
            
            DatHFpredIt1 = np.genfromtxt(f'{in_files}/HFGPR_HFpred_{noisetxt}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-{acqname}_Nhf{N_hf}_HFit{bb}.csv', skip_header=1, delimiter=',')
            Prange_predHF = DatHFpredIt1[:,0]
            mean_high_gp_model = DatHFpredIt1[:,1]
            var_high_gp_model = DatHFpredIt1[:,2]
            U95_high_gp_model = 1.96*np.sqrt(var_high_gp_model)
            
            ###### Plot HF
            plt.figure(figsize=(fsize_h,fsize_v))
            plt.yscale('log')
            plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS')
            plt.plot(v_train_h1,p_train_h1, color='r', marker='d',linestyle='None', label=f'{bb} HF samples')
            plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction')
            plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                              (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction U95%')
            plt.plot([0,15],[pexp,pexp],':',color='black')
            plt.text(14,pexp*1.4,'$P_{exp}$')
            plt.ylabel('P(C\'\'$\geq$c) [-]')
            plt.xlabel('c [m]')
            plt.ylim([10**(-5),1])
            plt.xlim([0,15])
            plt.legend()
            plt.savefig(f'{out_figs}/Prediction_HFGPR_{noisetxt}_Set-{HFkern}_{sampopt_txt}_{acqname}_it{bb}_intermediate.pdf',bbox_inches='tight')
    
            
            ############### READ AND PLOT INTERMEDIATE MF-GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
            
            # MF-GPR (4x)
            DatLFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFTrainData_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{bb}.csv', skip_header=1, delimiter=',')
            v_train_l2 = DatLFtrainIt2[:,0]
            p_train_l2 = DatLFtrainIt2[:,1]
            
            DatHFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFTrainData_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{bb}.csv', skip_header=1, delimiter=',')
            v_train_h2 = DatHFtrainIt2[:,0]
            p_train_h2 = DatHFtrainIt2[:,1]
            
            DatLFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFpred_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{bb}.csv', skip_header=1, delimiter=',')
            Prange_predLF = DatLFpredIt2[:,0]
            mean_low_mfgp_model = DatLFpredIt2[:,1]
            var_low_mfgp_model = DatLFpredIt2[:,2]
            U95_low_mfgp_model = 1.96*np.sqrt(var_low_mfgp_model)
            
            DatHFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFpred_{noisetxt}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MFit{bb}.csv', skip_header=1, delimiter=',')
            Prange_predHF = DatHFpredIt2[:,0]
            mean_high_mfgp_model = DatHFpredIt2[:,1]
            var_high_mfgp_model = DatHFpredIt2[:,2]
            U95_high_mfgp_model = 1.96*np.sqrt(var_high_mfgp_model)
            
            ###### Plot MF
            plt.figure(figsize=(fsize_h,fsize_v))
            plt.yscale('log')
            plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS')
            plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS')
            plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
            plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None', label=f'{bb} HF samples')
            plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',label='MF-GPR predictions')
            plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                              (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
            plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF (label not used)')
            plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                              (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
            plt.plot([0,15],[pexp,pexp],':',color='black')
            plt.text(14,pexp*1.4,'$P_{exp}$')
            plt.ylabel('P(C\'\'$\geq$c) [-]')
            plt.xlabel('c [m]')
            plt.ylim([10**(-5),1])
            plt.xlim([0,15])
            handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
            order = [0,1,2,3,4,5] #specify order of items in legend
            plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
            plt.savefig(f'{out_figs}/Prediction_MFGPR_{noisetxt}_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_it{bb}_intermediate.pdf',bbox_inches='tight')
                
            
            
# Finalise and save plot
plt.figure(10000)
if annon == 1:
    plt.savefig(f'{out_figs}/Fig12_Convergence_AllCases_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_converged_ann.pdf',bbox_inches='tight')
else:
    plt.savefig(f'{out_figs}/Fig12_Convergence_AllCases_GPR-{sampopt_txt}_MFGPR-{sampopt_txt}_converged.pdf',bbox_inches='tight')

            
            
            
    
    
