# -*- coding: utf-8 -*-
"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen

"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import glob
import os
import re

# Add some other functions
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.convergenceAS import stopping_crit_AS

# Close all figures
plt.close('all')



#############
# START INPUT
#############

# Plotsize
fsize_h = 7
fsize_v = 5

# Number of seeds to include in plot
seedmaxplot = 201
seedstep = 20

# Kernel name single-fidelity HF-GPR (Mat32 preferred for all)
HFkern = 'Mat32'
MFkern1 = 'Mat32' 
MFkern2 = 'Mat32'

# Acquisition function.
acqname = 'USMV'

# Folder name where the output of the Adaptive Screening procedure is saved.
# In the default case this is the same folder as where we write the output, but
# this can be defined differently if preferred.
in_files = 'out_files'

# Folder names in which the CSV files generated by Adaptive Screening were stored, 
# and name of the folder in which new figures will be stored. Defined relative 
# to the present folder, and the folder with this name needs to exist.
out_figs = 'out_figs'

# Corresponding line styles and colors in the plots
stylesgpr = '-.';     colsgpr = 'green'  # matching distribution plots
stylesmfgpr = '--';   colsmfgpr = 'blue' # matching distribution plots

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'B'

# Spectral properties basis wave system (for ref. Forristall and Pexp)
dur = 3600 # target duration
Hs = 10 # sign wave height
Tp = 11 # peak wave period
wd = 30 # water depth
Vs = 0 # current speed or vessel speed
T1 = Tp/1.198 # estimated from relation for JONSWAP

# Forristall
VrangeRF = np.linspace(0, 15, 101)[:, None]

# True HF MCS result (used realisation)
MPM_mcs = 10.27
MPMU95_mcs = 0.13

# Reference conventional result case
MPM_conv = 10.58
MPMU95_conv = 0.88
HFtime_conv = 5

# SQI values for the different indicator quality cases (sqi), number of HF 
# samples for convergence with GPR (nev_gpr) and number of HF samples for 
# convergence with MF-GPR (nev_mfgpr). 
# Cases: pi = PerfInd, gi = GoodInd, mi = MediumInd and bi = BadInd
sqi_pi = 1;    nev_gpr_pi = 23;   nev_mfgpr_pi = 22
sqi_gi = 2.6;  nev_gpr_gi = 87;   nev_mfgpr_gi = 91
sqi_mi = 3.7;  nev_gpr_mi = 122;  nev_mfgpr_mi = 89
sqi_bi = 4.9;  nev_gpr_bi = 126;  nev_mfgpr_bi = 102


#############
# END INPUT
#############



# Check wether output directory exist, if not make it
isExist1 = os.path.exists(out_figs)
if not isExist1:
   os.makedirs(out_figs)

# Initial sample options (see Adaptive Screening)
if sampopt == 'A':
    sampopt_txt = 'A_LFsubsamp'; N_hf = 2
elif sampopt == 'B':
    sampopt_txt = 'B_LFfit'; N_hf = 2
elif sampopt == 'C':
    sampopt_txt = 'C_LFisHFfit'; N_hf = 2

# Derive target probability of exceedance, using the average number of wave encounters in target duration
ne_exp = dur/Tp 
pexp = 1/ne_exp

# Calc Forristall ref distribution
p_For,dum2 = Forr(VrangeRF,Hs,T1,wd,Vs)

# Initialise
conv_mpm_gpr = np.nan*np.zeros(4)
conv_mpmu95_gpr = np.nan*np.zeros(4)
conv_mpm_mfgpr = np.nan*np.zeros(4)
conv_mpmu95_mfgpr = np.nan*np.zeros(4)

sqi_all = []
mpm_gpr_all = []
mpm_mfgpr_all = []
mpmu95_gpr_all = []
mpmu95_mfgpr_all = []
time_gpr_low = []
time_mfgpr_low =[]
time_gpr_hig = []
time_mfgpr_hig = []
nev_gpr_all = []
nev_mfgpr_all = []

# Lower and upper limit for event duration CFD [s]
tlow = 10
thigh = 50

# Loop over conditions (exclude the perfect indicator for now)
for nn in range(4):

    # Noise texts (figure titles and names)
    if nn == 0:
        indcase = 'PerfInd'
        sqi = sqi_pi
        nev_gpr = nev_gpr_pi
        nev_mfgpr = nev_mfgpr_pi
    elif nn == 1:
        indcase = 'GoodInd'
        sqi = sqi_gi
        nev_gpr = nev_gpr_gi
        nev_mfgpr = nev_mfgpr_gi
    elif nn == 2:
        indcase = 'MediumInd'
        sqi = sqi_mi
        nev_gpr = nev_gpr_mi
        nev_mfgpr = nev_mfgpr_mi
    elif nn == 3:
        indcase = 'BadInd'
        sqi = sqi_bi
        nev_gpr = nev_gpr_bi
        nev_mfgpr = nev_mfgpr_bi
    
    # Derive range in required duration [h]
    HFtime_gpr = [nev_gpr*tlow/3600,nev_gpr*thigh/3600]
    HFtime_mfgpr = [nev_mfgpr*tlow/3600,nev_mfgpr*thigh/3600]
    
    #### Find last available iteration (in this case probably #200, as we later interpolate in this database using a convergence criterion)

    # List all files in directory with the proper name, then select last available iteration
    # Getting the list of files/directories in the specified path, filtering the list to exclude the directory names 
    files_hf = list(filter(os.path.isfile, glob.glob(f'{in_files}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it*.csv')))
    files_mf = list(filter(os.path.isfile, glob.glob(f'{in_files}/Conv_MF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it*.csv')))
  
    # Find last available number of iterations
    ass_count_hf = np.array([],dtype=int)
    for file_i in files_hf:
        temp = int(re.search(r'\d+', file_i[-8:-1]).group())
        ass_count_hf = np.append(ass_count_hf,temp)
    count_hf_all = np.max(ass_count_hf)   
    del(temp)

    ass_count_mf = np.array([],dtype=int)
    for file_i in files_mf:
        temp = int(re.search(r'\d+', file_i[-8:-1]).group())
        ass_count_mf = np.append(ass_count_mf,temp)
    count_mf_all = np.max(ass_count_mf)    
    del(temp)
  
    print(f'{indcase} with HF-GPR and {acqname}: available are {count_hf_all} iterations ({count_hf_all+N_hf} samples)')
    print(f'{indcase} with MF-GPR and {acqname}: available are {count_mf_all} iterations ({count_mf_all+N_hf} samples)')
    
    
    #### Load data for this last iteration and plot convergence

    # Load data HF-GPR
    Dat_HF_GPR_all = pd.read_csv(f'{in_files}/Conv_HF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it{count_hf_all}.csv', delimiter=',')
    Dat_MF_MFGPR_all = pd.read_csv(f'{in_files}/Conv_MF-GPR_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_{acqname}_Nhf{N_hf}_it{count_mf_all}.csv', delimiter=',') # foutje in kernel names (Hf ipv 2xMF), maar al in andere file gemaakt.
   
    # Number of samples! (not iterations)
    noit_GPR = Dat_HF_GPR_all['NoIt'].to_numpy()
    noit_MFGPR = Dat_MF_MFGPR_all['NoIt'].to_numpy()

    # Actual MPM values
    mpms_GPR = Dat_HF_GPR_all['MPM_HFGPR'].to_numpy()
    mpmsU95_GPR = Dat_HF_GPR_all['U95_MPM_HFGPR'].to_numpy()
    
    mpms_MFGPR = Dat_MF_MFGPR_all['MPM_MFGPR'].to_numpy()
    mpmsU95_MFGPR = Dat_MF_MFGPR_all['U95_MPM_MFGPR'].to_numpy()
   
    
    #### Apply stopping criterion to determine at which iteration we are converged
    
    # File name start of CSV with predicted data per iteration (exclude "itXX.csv", will be added later)
    fn_preds_GPR = f'{in_files}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{sampopt_txt}_HFacq-{acqname}_Nhf{N_hf}_HF'
    casename_GPR = f'{indcase} - GPR - {acqname}'
    min_it1 = 10
    convit1 = stopping_crit_AS(noit_GPR,mpms_GPR,mpmsU95_GPR,fn_preds_GPR,casename_GPR,pexp,min_it1,ploton=False)
    
    fn_preds_MFGPR = f'{in_files}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{sampopt_txt}_MFacq-{acqname}_Nhf{N_hf}_MF'
    casename_MFGPR = f'{indcase} - MF-GPR - {acqname}'
    min_it2 = 10
    convit2 = stopping_crit_AS(noit_MFGPR,mpms_MFGPR,mpmsU95_MFGPR,fn_preds_MFGPR,casename_MFGPR,pexp,min_it2,ploton=False)
         
    #### CALCULATE AND PRINT MPM VALUES AND U95% FOR CONVERGED RESULTS
    conv_mpm_gpr = mpms_GPR[convit1]
    conv_mpmu95_gpr = mpmsU95_GPR[convit1]
    conv_mpm_mfgpr = mpms_MFGPR[convit2]
    conv_mpmu95_mfgpr = mpmsU95_MFGPR[convit2]
        
    #### Save data as a function of SQI        
    sqi_all = np.append(sqi_all,sqi)
    
    mpm_gpr_all = np.append(mpm_gpr_all,conv_mpm_gpr)
    mpm_mfgpr_all = np.append(mpm_mfgpr_all,conv_mpm_mfgpr)
    mpmu95_gpr_all = np.append(mpmu95_gpr_all,conv_mpmu95_gpr)
    mpmu95_mfgpr_all = np.append(mpmu95_mfgpr_all,conv_mpmu95_mfgpr)
    
    nev_gpr_all = np.append(nev_gpr_all,nev_gpr)
    nev_mfgpr_all = np.append(nev_mfgpr_all,nev_mfgpr)
    
    time_gpr_low = np.append(time_gpr_low,HFtime_gpr[0])
    time_mfgpr_low = np.append(time_mfgpr_low,HFtime_mfgpr[0])
    time_gpr_hig = np.append(time_gpr_hig,HFtime_gpr[1])
    time_mfgpr_hig = np.append(time_mfgpr_hig,HFtime_mfgpr[1])
            

############### Plot results as a function of SQI

y_ticks1 = np.arange(8,15,0.5)
y_ticks2 = np.arange(0,201,40)
y_ticks3 = np.arange(0,20,2)
xticks = np.arange(0,6,1)

# Plot excluding annotation
plt.figure(figsize=(7.5,6.5))
lw = 0.7

ax1 = plt.subplot(3,1,1)
plt.plot(sqi_all,mpm_gpr_all,marker='o',linestyle=stylesgpr,color=colsgpr,label='GPR (+U95%)',linewidth=lw)
plt.plot(sqi_all,mpm_mfgpr_all,marker='d',linestyle=stylesmfgpr,color=colsmfgpr,label='MF-GPR (+U95%)',linewidth=lw)
plt.errorbar(sqi_all,mpm_gpr_all,yerr=mpmu95_gpr_all/2,color=colsgpr,capsize=3,linestyle='none')
plt.errorbar(sqi_all,mpm_mfgpr_all,yerr=mpmu95_mfgpr_all/2,color=colsgpr,capsize=3,linestyle='none')
plt.plot(0,MPM_conv,marker='o',linestyle='none',color='darkorange',label='Conventional (+U95%)')
plt.errorbar(0,MPM_conv,yerr=MPMU95_conv/2,color='darkorange',capsize=3)
plt.plot([-1,5.5],[MPM_mcs,MPM_mcs],linestyle=':',color='grey')    
plt.text(0.26,MPM_mcs*0.975,'True MCS',color='grey')
plt.legend(loc='upper right',ncol=3)
plt.ylabel('$\widehat{C\'\'}$ [m]')
plt.xticks(ticks=xticks,labels=[])
plt.yticks(ticks=y_ticks1)
plt.ylim([9,12])
plt.xlim([-0.3,5.2])

ax2 = plt.subplot(3,1,2)
plt.plot(sqi_all,nev_gpr_all,marker='o',linestyle=stylesgpr,color=colsgpr,label='GPR',linewidth=lw)
plt.plot(sqi_all,nev_mfgpr_all,marker='d',linestyle=stylesmfgpr,color=colsmfgpr,label='MF-GPR',linewidth=lw)
plt.yticks(ticks=y_ticks2)
plt.ylabel('# required HF samples')
plt.xticks(ticks=xticks,labels=[])
plt.ylim([0,200])
plt.xlim([-0.3,5.2])
plt.legend(loc='upper right',ncol=2)

ax3 = plt.subplot(3,1,3)
plt.fill_between(sqi_all,time_gpr_low,time_gpr_hig,facecolor=colsgpr, alpha=0.15,label='GPR (forecast)')
plt.fill_between(sqi_all,time_mfgpr_low,time_mfgpr_hig,facecolor=colsmfgpr, alpha=0.15,label='MF-GPR (forecast)')
plt.plot(0,HFtime_conv,marker='o',linestyle='none',color='darkorange',label='Conventional')
plt.yticks(ticks=y_ticks3)
plt.xlabel('SQI [-]') 
plt.ylabel('Estimated time to \n simulate with HF tool [h]')
plt.xticks(ticks=xticks)
plt.ylim([0,8])
plt.xlim([-0.3,5.2])
plt.legend(loc='upper right',ncol=3)

plt.savefig(f'{out_figs}/Fig15_PredVsSQI_allkernsMat32_HF-{sampopt_txt}_MF-{sampopt_txt}_HFacq-{acqname}_MFacq-{acqname}_exclann.pdf',bbox_inches='tight')


