# -*- coding: utf-8 -*-
"""
Created on Fri Jan 21 11:16:08 2022

@author: smvanessen

This script contains FFT functions, specific for wave analysis. So spectrum to 
time traces, time traces to spectrum etc.

"""

# Required toolboxes and import syntaxes
import numpy as np
import matplotlib.pyplot as plt
import random 

def transform_spec2tt(om,spec,pha):
    
    '''
    This function uses the standard Matlab fft to translate a signal spectrum 
    to a time trace. 
    Input:    - Frequency axis [om] in <rad/s>
              - Spectrum [spec] in <unit^2.s/rad>
              - Phase components [pha] in <rad>, or seed number. If pha is a
                vector of the same size as the frequency axis, it is assumed
                to be the phase vector. If pha is a scalar, it is assumed to
                be a seed number, and a random seed vector is generated.
    Output:   - Time axis [time] in <s>
              - Signal time trace [signal] in <unit>
              - Spectral phases [pha] in <rad>, on input frequency axis.
                Always present in output, but only useful when input
                parameter [pha] is defined as a seed number. The 'explicit'
                phases in the output could be used for further processing.
    Reconstruction of time traces: sig-out = sum(amp.*cos(om*time+pha))
    
    Revision: - Original script, Mar 2021, SvE
              - Matlab > Python, Jan 2022, SvE
    '''
    
    # Generate phase vector if a random seed is provided.
    if np.isscalar(pha):
        seed1 = pha
        # rand('state',seed)
        random.seed(seed1)
        pha = np.zeros(np.size(om))
        irange = np.arange(0,len(om),1)
        for ii in irange:
            pha[ii] = random.uniform(0,1)*2*np.pi
        
    # Check that all entries are non-nan
    if any(np.isnan(om)):     
        raise ValueError("Frequency axis has NaN entry")
    if any(np.isnan(spec)):   
        raise ValueError("Spectrum has NaN entry")
    if any(np.isnan(pha)):    
        raise ValueError("Phases have NaN entry")
    
    # Check dimensions
    if np.size(spec) != np.size(pha):
        raise ValueError("Dimension spec and pha must be the same")
    if len(spec) != len(om):
        raise ValueError("First dim. of spec must equal length(om)")
    
    # Number of components
    N = len(om)
    
    # Frequency step
    dom = (om[-1] - om[0])/(len(om) - 1)
    
    # Amplitudes
    amp = np.sqrt(2*spec*dom)
    
    # DFT
    # Equivalent to: DFT = N*amp*(np.cos(pha)+np.sqrt(-1j)*np.sin(pha))
    DFT = N*amp*np.exp(1j*pha)
    
    # Time step
    dt = 2*np.pi/dom/N
    
    # Time axis
    # time = 0:dt:(N-1)*dt
    # time = np.arange(0,(N-1)*dt,dt)
    time = np.arange(0,N*dt,dt)
    
    # s = len(DFT)
    # DFTL  = [DFT; zeros([N-length(DFT) s(2:end)])]
    # DFTL = np.concatenate((DFT,np.zeros((N-len(DFT),s))))
    # signal = np.real(np.fft.ifft(DFTL))
    signal = np.real(np.fft.ifft(DFT))
    
    return time,signal,pha


# ADD tt2spec