# -*- coding: utf-8 -*-
"""
Created on Thu Nov 23 10:16:10 2023

@author: smvanessen
"""

# Import standard toolboxes
import numpy as np
import matplotlib.pyplot as plt

# Color functions
from matplotlib import cm
from matplotlib.colors import Normalize



def stopping_crit_AS(noit,MPMs,MPMsU95,filename_preds_per_it,casename,pexp,minnoit,ploton):
    """
    Default version, with acceptance criterion, mean absolute difference criterion 
    for distributions and COV criterion for MPM. MODIFY DESCRIPTION!
    
    This function evaluates the stopping criteria for Adaptive Screening, as 
    discussed in the paper. This includes a monotononicity acceptance criterion, 
    a criterion for the Kolmogorov-Smirnov statistic value of consecutive 
    iterations (averaged over 25 iterations) and a criterion for the average COV 
    of the most probable maximum (MPM) over 20 iterations. NB. these numbers of 
    iterations and the criteria epsilon1 and epsilon2 are hardcoded here, as
    discussed in the 2nd-order wave paper. For other problems, these may have 
    to be changed.
    
    Parameters
    ----------
    noit : numpy array
        Number of included samples for all iterations
    MPMs : numpy array
        MPM values per iteration
    MPMsU95 : numpy array
        U95% of this MPM value per iteration
    filename_preds_per_it : string
        Filename of the CSV file that contains the full prediction for each iteration. 
        Generated by 'MF_GPR_Lin2ndOrd_Nseeds16_adaptive', full filename including 
        path, but excluding the extension "itXX.csv". Example: for file 
        'files/HFGPR_HFpred_NoNoise_Set-Mat32-nf_AdaptiveInit5LFsubsamp_HFacq-US_Nhf2_HFit20.csv',
        this variable would be 'files/HFGPR_HFpred_NoNoise_Set-Mat32-nf_AdaptiveInit5LFsubsamp_HFacq-US_Nhf2_HF'
    casename : string
        Name to give the output in the figures (example; 'NoNoise case - MF-GPR with US'). 
        Do not use semicolons.
    pexp: scalar
        Target probability of exceedance (used here as reference line in the distribution plot).
    minnoit; scalar
        Mininum number of iterations; convergence before this number is ignored
    ploton : boolean
        Plot switch. True = plots, false = no plots.
    

    Returns
    -------
    it_conv : scalar
        Index of <noit> at which the results are converged according to the set stopping criteria.

    """
    
    ###### Hardcoded limits of the criterion that may have to be changed in another application
    
    # Number of last iterations for average max. abs. difference criterion (if less are available, we take the average over all available)
    N_mae = 20
    
    # Number of last iterations for average COV of the MPM (if less are available, we take the average over all available)
    N_cov = 20
    
    # Limit for maximum absolute difference (in this case in [m] wave height)
    lim_MAE_conv = 0.03

    # Limit for COV of the MPM / quantile convergence criterion over the last 10 iterations (N_cov)
    # Unit is related to percentage of the mean (absolute) wave height
    lim_cov_conv = 0.00098

    
    ###### End hardcoded limits
    
      
    # Case name exluding spaces, for file names
    casename_f = casename.replace(' ','_')
        
    # # Max. number of available iterations
    itmax = len(noit)
    
    # Define the colormap Viridis and normalize it to the data range
    # cmap = plt.get_cmap('Blues')  # You can use any other colormap
    # norm = Normalize(vmin=-itmax/2, vmax=itmax)
    cmap = plt.get_cmap('jet')  # You can use any other colormap
    norm = Normalize(vmin=0, vmax=itmax)
    
    # Create a ScalarMappable object to map data values to colors
    sm = cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([]) 
   
    # Initialise assembly of mean predictions
    ass_meanpred = []
    
    # Initialise rejection test for monotonicity
    RejectIt = np.ones(itmax)
         
    # initialise MAE 
    D_MAE = np.nan*np.zeros(itmax)
    D_MAE_av = np.nan*np.zeros(itmax)
    
    # Initialise mean, standard deviation and coefficient of variation of MPM
    stdconv_mpm = np.nan*np.zeros(itmax)
    meanconv_mpm = np.nan*np.zeros(itmax)
    cov_mpm = np.nan*np.zeros(itmax)
    cov_mpm_last = np.nan*np.zeros(itmax)
    
    # Initialise final criterion that combines 1,2,3 (starts at 0)
    crit_all = np.ones(itmax)

    # Initialise figure
    if ploton == True:
        fig1 = plt.figure()
    
    # Loop over number of iterations
    for ii in range(itmax):
               
        # Color range for increasing it. number
        colnow = [0,0,0.5+(ii/itmax)/2]
        
        # Load data - full predicted HF distributions
        preddata = np.genfromtxt(f'{filename_preds_per_it}it{ii}.csv', skip_header=1, delimiter=',')
 
        # Extract PoE range and mean predictions (HF version MF-GPR at same columns as GPR predictions)
        Prange_pred = preddata[:,0]
        mean_pred = preddata[:,1]
        
        # MPM and MPM U95 values up to the present iteration
        mpms_now = MPMs[0:ii]
        u95mpms_now = MPMsU95[0:ii]

        
        ##################  
        # Crit1: Reject non-monotonic results 
        ################## 
          
        # Check whether prediction in present iteration is monotonically decreasing, 
        # and that there is a difference between top and down of more than 1% of top value
        eps1 = 0.01*mean_pred[0]
        if all(mean_pred[i] > mean_pred[i+1] for i in range(len(mean_pred)-1)):
            # Yes, monotonically decreasing
            if mean_pred[0]-mean_pred[-1] > eps1:
                # Mon. decreasing, but difference top-down big enough
                RejectIt[ii] = 0
            else:
                # Mon. decreasing, but difference top-down too small
                RejectIt[ii] = 1
        else:
            # No, not monotonically decreasing
            RejectIt[ii] = 1
        
        ################## End crit1

        # Plot all distributions on top of each other in the open figure
        if ploton == True:
            plt.figure(fig1)
            if RejectIt[ii] == 1:
                plt.plot(mean_pred,Prange_pred,color='grey')
                plt.text(mean_pred[-1],Prange_pred[-1],f'{ii}',color='grey')                    
            else:
                if np.mod(ii,20) == 0:
                    plt.plot(mean_pred,Prange_pred,color=sm.to_rgba(ii),label=f'Iteration {ii}')
                else:
                    plt.plot(mean_pred,Prange_pred,color=sm.to_rgba(ii))
                plt.text(mean_pred[-1],Prange_pred[-1],f'{ii}',color=sm.to_rgba(ii))
                    
        # Save these data to use in convergence assessment
        ass_meanpred.append(mean_pred)
        
        ##################  
        # Crit2: Calculate maximum absolute error
        ##################     
                            
        # Loop starts from 2 available iterations.
        if ii > 0:
            
            # DNR present and previous iteration
            pred1 = ass_meanpred[ii-1]
            pred2 = ass_meanpred[ii]
            
            # max absolute difference
            # MAE = np.mean(np.abs(pred1-pred2))
            MAE = np.max(np.abs(pred1-pred2))
            # MAE = np.sqrt(np.mean((pred1-pred2)**2))
            
            # Assemble data
            D_MAE[ii] = MAE
        
        
        # Very erratic; calculate the MAE over the last N_mae iterations
        # Start at > 1 because D_MAE[:0] = [] and D_MAE[:1] = D_MAE[0] = nan
        if ii > 1:
            if ii <= N_mae-1:
                # "Burn-in": discard portion of simulation results to ensure steady state (or treat them differently)
                D_mae_t = D_MAE[:ii][np.isfinite(D_MAE[:ii])]
            elif ii > N_mae-1:
                D_mae_t = D_MAE[ii-N_mae:ii][np.isfinite(D_MAE[ii-N_mae:ii])]
            D_MAE_av[ii] = sum(D_mae_t)/len(D_mae_t)
          
       
       ################## End crit2
        
        
        ##################  
        # Crit3: Calculate standard deviation and COV of MPM over existing iterations
        ##################     
        
        # Use the `moving average' standard deviation, divided by sqrt(number of samples)
        # This stems from the central limit theorem, saying that the mean of 
        # a number of samples from any distribution is normally distributed.
        # Here we don't have the mean and we don't have random samples, but 
        # it may still work.
                  
        # Sample standard deviation
        # stdconv_mpm[ii] = np.std(mpms_now)/np.sqrt(ii)
        stdconv_mpm[ii] = np.std(mpms_now)
        
        # Sample mean
        meanconv_mpm[ii] = np.mean(mpms_now)  
        
        # Coefficient of variation up to present iteration
        cov_mpm[ii] = stdconv_mpm[ii]/meanconv_mpm[ii]

        # Very erratic; calculate the COV over the last N_cov iterations
        # Start at > 0 because mpms_now[:0] = []
        if ii > 1:
            if ii <= N_cov-1:
                # "Burn-in": discard portion of simulation results to ensure steady state (or treat them differently)
                mpms_t = mpms_now[:ii][np.isfinite(mpms_now[:ii])]
            elif ii > N_cov-1:
                mpms_t = mpms_now[ii-N_cov:ii][np.isfinite(mpms_now[ii-N_cov:ii])]
            # cov_mpm_last[ii] = (np.std(mpms_t)/np.sqrt(len(mpms_t)))/np.mean(mpms_t)  
            cov_mpm_last[ii] = np.std(mpms_t)/np.mean(mpms_t)      
        
        ################# End crit3

        ###########
        # One criterion that combines them all
        ###########
        
        # Start with minimum number of iterations
        crit_all[:minnoit] = 0
            
        # First check monotonicity criterion
        if RejectIt[ii] == 1:
            crit_all[ii] = 0
        
        # Check the mean absolute error convergence criterion
        if D_MAE_av[ii] > lim_MAE_conv:
            crit_all[ii] = 0
        elif np.isnan(D_MAE_av[ii]):
            crit_all[ii] = 0  
        
        # Check the MPM / quantile COV convergence criterion
        if cov_mpm_last[ii] > lim_cov_conv:
            crit_all[ii] = 0
        elif np.isnan(cov_mpm_last[ii]):
            crit_all[ii] = 0
      
        
      
    #### End of loop over iterations
    
    
    # Finish figure of all distributions
    if ploton == True:
        plt.figure(fig1)
        plt.yscale('log')
        plt.title(f'{casename} \n (grey = rejected iteration)')
        plt.ylabel('P(c$\geq$C) [-]')
        plt.xlabel('C [m]')
        plt.ylim([4*10**(-4),2*10**(-1)]) # room for legend
        plt.xlim([7.5,12.5])
        plt.legend(ncol=2)
        plt.plot([0,15],[pexp,pexp],':',color='black')
        plt.text(12.1,pexp*1.2,'$P_{exp}$')
        plt.savefig(f'figures3/AllDistrForIts_{casename_f}.pdf',bbox_inches='tight')

    # First converged iteration number
    it_conv = np.argmax(crit_all>0)
    print(f'Case {casename}: converged at {it_conv} iterations.')
    
    
    # Plot of convergence and criteria
    if ploton == True:
    
        # Location of text 'Criterion'
        x_loc_crit = 175
        x_ticks = np.arange(0,202,25)

        # Max. number of samples (plot limit)
        seedmaxplot = noit[-1]+1
        
        # Colors and linestyles
        linestyles = ['-','--','-.',':']
        cols = ['royalblue','indianred','mediumseagreen','orange']
                
        # Rejection mask
        nanmask_rej = np.where(RejectIt==1,np.nan,1)
        
        # --------- ALTERNATIVE PLOT NUMBER 2
        plt.figure(figsize=(6.1,6))
        
        # MPM    
        ax1 = plt.subplot(3,1,1)
        ax1.spines['top'].set_visible(False) # Only removes box lines
        ax1.spines['right'].set_visible(False)
        plt.ylabel('$\widehat{C\'\'}$ [m]')
        plt.plot(noit[:it_conv-1],MPMs[:it_conv-1],linestyle=linestyles[0],color=cols[0])  
        plt.plot(noit[it_conv:],MPMs[it_conv:],linestyle=linestyles[0],color=cols[0],alpha=0.2)  
        plt.plot(noit[it_conv],MPMs[it_conv],linestyle='none',color=cols[0],marker = 'x',label='Converged result') 
        plt.ylim([10,10.6])
        plt.xlim([1,seedmaxplot])
        plt.xticks(ticks=x_ticks,labels=[])
        plt.legend()
        plt.title(f'{casename} - Convergence')
       
        # Max abs difference criterion 2 (averaged over number of iterations)
        ax2 = plt.subplot(3,1,2)
        ax2.spines['top'].set_visible(False) # Only removes box lines
        ax2.spines['right'].set_visible(False)
        plt.ylabel(r'$\overline{E}_{%i}$ [m]' % (N_mae))
        plt.plot(noit,D_MAE_av*nanmask_rej,linestyle=linestyles[0],color=cols[0])
        plt.xlim([1,seedmaxplot])
        plt.plot([0,seedmaxplot],[lim_MAE_conv,lim_MAE_conv],linestyle=':',color='grey',linewidth=0.8)
        plt.text(x_loc_crit,lim_MAE_conv*1.5,'Criterion',color='grey')
        plt.yscale('log')
        plt.ylim([10**(-3),10])
        plt.xticks(ticks=x_ticks,labels=[])
        
        # MPM COV criterion 3 (plus averaged over a number of iterations)
        ax3 = plt.subplot(3,1,3)
        ax3.spines['top'].set_visible(False) # Only removes box lines
        ax3.spines['right'].set_visible(False)
        plt.ylabel(r'$C_{%i}$ [-]' % (N_cov))
        plt.plot(noit,cov_mpm_last*nanmask_rej,linestyle=linestyles[0],color=cols[0])
        plt.xlim([1,seedmaxplot])
        plt.plot([0,seedmaxplot],[lim_cov_conv,lim_cov_conv],linestyle=':',color='grey',linewidth=0.8)
        plt.text(x_loc_crit,lim_cov_conv*1.5,'Criterion',color='grey')
        plt.yscale('log')
        plt.ylim([10**(-5),10**(-1)])
        plt.xticks(ticks=x_ticks)
        plt.xlabel('Number of samples $j$ [-]')
      
       
        plt.savefig(f'figures3/Convergence_stopcrit_{casename_f}.pdf',bbox_inches='tight')

         
    return it_conv
    
