# -*- coding: utf-8 -*-
"""
Created on Thu Jan 13 13:15:55 2022

@author: smvanessen

This file contains scripts concerning zero up-crossing peak and trough detection,
generation of distributions etc.

"""

# Required toolboxes and import syntaxes
import numpy as np
import matplotlib.pyplot as plt

def calc_zup(t,sig,ploton):

    ''' 
    This function calculates the zero up-crossings in an input time signal,
    the corresponding peaks and troughs and their indices. The function uses
    the first derivative of the input time traces to determine where the zero 
    up-crossings are (signal = 0 and derivative signal = positive). Note that 
    the detection will never be more accurate in time than the time step dt.
    Peaks and troughs are defined after the corresponding zero up-crossing
    (so peaks and troughs before the first zero up-crossing are disregarded).
    Peaks and troughs after the last zero up-crossing are also disregarded
    (as there is no next zero up-crossing to bound the interval in which a
    maximum and minimum can be found).
    
    --- Input ---
    t             Time axis [s]
    sig           Signal time trace [<U>]    
    ploton        Plot switch (1 = plot, 0 = no plot)
    --- Output ---
    t_peaks       Times of zero up-crossing peaks w.r.t. time axis [s]
    peaks         Zero up-crossing peaks [<U>] - the zero up-crossing peak is
                  defined to occur after the corresponding zero up-crossing.
    t_troughs     Times of zero up-crossing troughs w.r.t. time axis [s]
    troughs       Zero up-crossing troughs [<U>]- the zero up-crossing peak is
                  defined to occur after the corresponding zero up-crossing.
    t_zup         Times of zero up-crossings w.r.t. time axis [s]
    
    --- Revisions ---
    Mar 2021 SvE Original Matlab script
    Jan 2022 SvE Matlab > Python
    
    '''

    # Time step
    dt = (t[-1]-t[0])/(len(t)-1)

    # Find zero up-crossings (changes from negative to positive values)
    i_zup = np.where(np.diff(np.sign(sig))>0)
    i_zdown = np.where(np.diff(np.sign(sig))<0) # not in output
    i_zup = i_zup[0]
    i_zdown = i_zdown[0]
    # Check why the np.where function outputs a tuple instead of an array. This has to be corrected below before adding the 1 to all indices.

    # Make up-crossing times more precise by interpolating linearly between the two samples around it.
    v_zup_low = sig[i_zup]
    v_zup_high = sig[i_zup+1]

    t_zup_low = t[i_zup]
    t_zup_high = t[i_zup+1]
    
    t_zup = (-v_zup_low*t_zup_high + v_zup_high*t_zup_low)/(-v_zup_low + v_zup_high)
    
    # Same for down-crossing times
    v_zdown_low = sig[i_zdown]
    v_zdown_high = sig[i_zdown+1]
    
    t_zdown_low = t[i_zdown]
    t_zdown_high = t[i_zdown+1]
    
    t_zdown = (v_zdown_low*t_zdown_high - v_zdown_high*t_zdown_low)/(v_zdown_low - v_zdown_high)

    # Derivative plot (note shift of dt/2 to match input time signal)
    if ploton == 1:
        fig1 = plt.figure()
        ax1 = fig1.add_subplot(1,1,1)
        pl1 = ax1.plot(t,sig,color='blue',label='sig')
        pl2 = ax1.plot(t,np.sign(sig),color='red',label='sign(sig)')
        pl3 = ax1.plot(t[0:-1]+dt/2,np.diff(np.sign(sig)),color='green',label='diff(sign(sig))')
        pl4 = ax1.plot(t_zup,np.zeros(np.shape(i_zup)),marker='o',color='black',label='zero up-crossings',linestyle = 'None')
        pl5 = ax1.plot(t_zdown,np.zeros(np.shape(i_zdown)),marker='x',color='purple',label='zero down-crossings',linestyle = 'None')
        # plt.legend(handles=[pl1[0],pl2[0],pl3[0],pl4[0],pl5[0]],labels=['sig','sign(sig)','diff(sign(sig))','zero up-crossings','zero down-crossings'])
        plt.legend(handles=[pl1[0],pl2[0],pl3[0],pl4[0],pl5[0]])
        plt.xlabel('time [s]')
        plt.ylabel('wave [m]')
        plt.title('Identification of zero up- and down-crossings')
        plt.show()

    # Find corresponding peaks & peak times (peak N occurs after up-crossing N)
    # And similarly the troughs & trough times
    # This version uses 'Matlab indexing'
    peaks = np.nan*np.ones(np.shape(t_zup))
    t_peaks = np.nan*np.ones(np.shape(t_zup))
    troughs = np.nan*np.ones(np.shape(t_zup))
    t_troughs = np.nan*np.ones(np.shape(t_zup))
    irange1 = np.arange(0,len(t_zup)-1,1)
    for ii in irange1:
        inds1 = np.nonzero((t >= t_zup[ii]) & (t < t_zup[ii+1]) )
       
        # Peaks
        peaks[ii] = np.max(sig[inds1])
        locind1 = np.argmax(sig[inds1])
        # t_peaks[ii] = (locind1)*dt + t[i_zup[ii]]
        t_peaks[ii] = (locind1+1)*dt + t[i_zup[ii]]
        
        # Troughs
        troughs[ii] = np.min(sig[inds1])
        locind2 = np.argmin(sig[inds1])
        # t_troughs[ii] = (locind2)*dt + t[i_zup[ii]]
        t_troughs[ii] = (locind2+1)*dt + t[i_zup[ii]]
        
    # =============================================================================
    #     # This version uses 'Python indexing' (WERKT NOG NIET!)
    #     peaks2 = np.nan*np.ones(np.shape(t_zup))
    #     t_peaks2 = np.nan*np.ones(np.shape(t_zup))
    #     ind2 = 0
    #     for jj in i_zup[0:-1]:
    #         inds3 = np.nonzero((t >= t[jj]) & (t < t[jj+1]) ) # This does not work because jj+1 is not equal to the next i_zup value.
    #         peaks2[ind2] = np.max(sig[inds3])
    #         locind = np.argmax(sig[inds3])
    #         t_peaks2[ind2] = (locind-1)*dt + t[jj]
    #         ind2 = ind2+1
    # =============================================================================
    
    # Add to plot
    if ploton == 1:
        pl6 = ax1.plot(t_peaks,peaks,marker='+',color='grey',label='peaks',linestyle = 'None')
        pl7 = ax1.plot(t_troughs,troughs,marker='*',color='yellow',label='troughs',linestyle = 'None')
        plt.legend(handles=[pl1[0],pl2[0],pl3[0],pl4[0],pl5[0],pl6[0],pl7[0]])
        plt.title('Identification of zero up-crossings, peaks and troughs')
   
    return t_peaks,peaks,t_troughs,troughs,t_zup



def calc_distribution(signal):
    
    '''
    This function calculates the distribution of the input signal. So if the 
    input signal is a time trace, the output is a sample distribution. If the 
    input signal is the peaks of a time trace, the output is a peak distribution.
    -----
    Input:
    signal    = Signal to calculate distribution for (can be samples, peaks, troughs, etc.)
    
    
    -----
    Output:
    vals      = Values for which the distribution is calculated.
    probs     = Corresponding exceedance probabilities P(x >= vals) based on 
              the input data. The largest sample has P = 1/N, where N is the 
              number of samples.
    -----
    Revision: - Original script, Mar 2021, SvE
              - Matlab > Python, Jan 2022, SvE
    -----
    '''
    
    # Default values for x-axis of distribution (ascending order)
    vals = np.sort(signal)
        
    # Number of samples
    N = len(signal)
    
    # Initialise probability array
    probs = np.zeros(np.size(vals))
    probs2 = np.zeros(np.size(vals))
    
    # Calculate probability that signal sample >= vals(ii)
    irange2 = np.arange(0,len(vals),1)
    for ii in irange2:
        probs[ii] = np.sum(signal >= vals[ii])/N
    
    # Test python-style loop
    ct = 0
    for jj in vals:
        probs2[ct] = np.sum(signal >= jj)/N
        ct = ct + 1

    # Maybe add later: optional input argument vals
    # vals      = Values for which the distribution is calculated. If it is not
    #           specified, the default is used: the input values, sorted low to
    #           high. NB possible alternative: evenly spaced values, with the 
    #           same number of steps as input signal samples, between 
    #           min(signal) and max(signal).            
    
    return vals,probs


def MPM_from_DNR(V_DNR,P_DNR,np_exp):
    # Calculates MPM from DNR distribution (all crests). See papers below.
    # Assumes that the largest peak has probability 1/n_tot, where n_tot is the 
    # total number of peaks in all available realisations combined.
    # Input: 
    # - V_DNR = value range DNR
    # - P_DNR = exceedance probability DNR
    # - np_exp = average number of peaks in target exposure duration
    # Output:
    # - MPM = most probable maximum value for the target exposure duration
    # Refs:
    # - Van Essen, S. M., Scharnke, J. and Seyffert, H. C. (2023). Required test 
    #   durations for converged short-term wave and impact extreme value
    #   statistics - Part 1: ferry dataset. Marine Structures (Under Review).
    # - Scharnke, J., Van Essen, S. M. and Seyffert, H. C. (2023). Required test 
    #   durations for converged short-term wave and impact extreme value
    #   statistics - Part 2: deck box dataset. Marine Structures (Under Review)
    
    # Exposure duration exceedance probability level
    Pexp = 1/np_exp
    
    # If array is not 0D, squeeze
    # And flip such that P is monotonically increasing
    P_DNR0 = np.flip(np.squeeze(P_DNR))
    V_DNR0 = np.flip(np.squeeze(V_DNR))
    
    # Remove NaN values in V_DNR0
    nanmask = ~np.isnan(V_DNR0)
    P_DNR0_exclnan = P_DNR0[nanmask]
    V_DNR0_exclnan = V_DNR0[nanmask]
 
    # Interpolate
    MPM = np.interp(Pexp,P_DNR0_exclnan,V_DNR0_exclnan)
    
    return MPM


def MPM_from_DEM(V_DEM,P_DEM):
    # Calculates MPM from DEM distribution (ensemble maxima). See papers below.  
    # Uses the Gaussian assumption that exceedance probability of MPM is 62.3%
    # Input: 
    # - V_DNR = value range DEM for target exposure duration
    # - P_DNR = exceedance probability DEM for target exposure duration
    # Output:
    # - MPM = most probable maximum value for the target exposure duration
    # Refs:
    # - Van Essen, S. M., Scharnke, J. and Seyffert, H. C. (2023). Required test 
    #   durations for converged short-term wave and impact extreme value
    #   statistics - Part 1: ferry dataset. Marine Structures (Under Review).
    # - Scharnke, J., Van Essen, S. M. and Seyffert, H. C. (2023). Required test 
    #   durations for converged short-term wave and impact extreme value
    #   statistics - Part 2: deck box dataset. Marine Structures (Under Review).
    
    P_DEM0 = np.flip(np.squeeze(P_DEM))
    V_DEM0 = np.flip(np.squeeze(V_DEM))
    
    MPM = np.interp(0.632,P_DEM0,V_DEM0)
    
    return MPM


def fit_weibull3_lsq(P_data, V_data, topPerc, maxit=None):
    # This function returns a 3-parameter Weibull fit for the input data.
    # Use Pweib = np.exp(-((Vweib - theta)/alpha)**beta) to reconstruct the fit 
    # for input values Vweib.
    
    
    # Give default number of iterations (20) if this is not specified in the input
    if maxit is None:
        maxit = 20

    # Keep only the less than topPerc% probability part
    topPerc = topPerc / 100
    nCrests = len(V_data)
    P_data = P_data[-int(np.floor(nCrests * topPerc)):]
    V_data = V_data[-int(np.floor(nCrests * topPerc)):]
    
    # Sort the values
    X = sorted(np.squeeze(V_data))
    Y = [x for _,x in sorted(zip(np.squeeze(V_data),np.squeeze(P_data)))]
    
    # startTheta and endTheta define the maximum range to look for theta
    startTheta = -X[1] + 2 * X[0]
    endTheta = startTheta + X[0] - X[-1]
    # print(f"Before loop: startTheta = {startTheta} and endTheta = {endTheta}") # debugging
    
    # R is the correlation coefficient which should be maximized
    R = np.zeros(20)
    cMatrix = []
    iMax = 0
    arrTheta = []
    theta = endTheta
    
    # Iterate to get the least squares theta
    # for j in range(maxit + 1):
    for j in range(maxit):

        if j == maxit:
            # print(f"Weibull-fitting least squares loop {j} of {maxit} (last extra iteration used to save data)")  # debugging
            arrTheta = arrTheta[iMax]
        else:
            # print(f"Weibull-fitting least squares loop {j} of {maxit}")  # debugging
            # Make array (descending)
            thetastep = (startTheta - endTheta)/19 # why 19?
            arrTheta = np.linspace(endTheta, endTheta+19*thetastep,20) # Corresponds to the 19
           
            # Reverse (ascending)
            arrTheta = arrTheta[::-1]
            
            # Print # elements
            # print(f"Length arrTheta = {len(arrTheta)}") # debugging

        # If this is empty, machine precision already reached, so stop
        if len(arrTheta) != 0:
            # Loop over the suggested theta's
            for i in range(len(arrTheta)):
                iTheta = arrTheta[i]
                N = len(X)
                F = np.log(X - iTheta)
                G = -np.log(-np.log(Y))
                aMatrix = np.array([[N, np.sum(F)], [np.sum(F), np.sum(F ** 2)]])
                bMatrix = np.array([np.sum(G), np.sum(F * G)])
                cMatrix = np.linalg.solve(aMatrix, bMatrix)
                b3 = np.sum(G ** 2)
                R[i] = np.sqrt((cMatrix[0] * bMatrix[0] + cMatrix[1] * bMatrix[1] - (bMatrix[0] ** 2 / aMatrix[0, 0])) / (b3 - (bMatrix[0] ** 2 / aMatrix[0, 0])))
        else:
            break
        iMax = np.argmax(R)
        theta = arrTheta[iMax]
        startTheta = arrTheta[max(iMax - 1, 0)]
        endTheta = arrTheta[min(iMax + 1, len(arrTheta) - 1)]
        # print(f"startTheta = {startTheta} and endTheta = {endTheta}")  # debugging

    alpha = np.exp(-cMatrix[0] / cMatrix[1])
    beta = -cMatrix[1]
    correlation = R[0]

    V_data_tp = V_data
    P_data_tp = P_data

    return V_data_tp, P_data_tp, theta, alpha, beta, correlation
    

