# -*- coding: utf-8 -*-
"""
Created on Fri Sep  8 11:45:32 2023

@author: smvanessen
"""

import numpy as np
# np.random.seed(42)

def subsample_binned(x,y,bins):
    
    """
    This function bins the input data in equal bins along the axis. 
    - If there are one or two points in a bin, take over those points
    - If there are more points in a bin, take over the minimum and maximum value
    - Appends the first and last point to make sure the same data range is used

    This allows to keep the original data in regions of low density, but 
    significantly reduce the amount of data to plot in regions of high density. 
    At the same time all the features are preserved with a sufficiently dense binning.
    
    Parameters:
    - x: 1D array, input data for the x-axis.
    - y: 1D array, input data for the y-axis.
    - bins: 1D array, bins for the x-axis.

    Returns:
    - xfilt: 1D array, filtered x values.
    - yfilt: 1D array, filtered y values.
    
    Source:
    https://stackoverflow.com/questions/53543782/how-can-i-subsample-an-array-according-to-its-density-remove-frequent-values
    
    """
    
    d = np.digitize(x, bins)
    xfilt = []
    yfilt = []
    for i in np.unique(d):
        xi = x[d == i]
        yi = y[d == i]
        if len(xi) <= 2:
            xfilt.extend(list(xi))
            yfilt.extend(list(yi))
        else:
            xfilt.extend([xi[np.argmax(yi)], xi[np.argmin(yi)]])
            yfilt.extend([yi.max(), yi.min()])
    # prepend/append first/last point if necessary
    if x[0] != xfilt[0]:
        xfilt = [x[0]] + xfilt
        yfilt = [y[0]] + yfilt
    if x[-1] != xfilt[-1]:
        xfilt.append(x[-1])
        yfilt.append(y[-1])
    sort = np.argsort(xfilt)
    return np.array(xfilt)[sort], np.array(yfilt)[sort]


def subsample_binned_2D(x, y, x_bins, y_bins):
    """
    
    This function bins the input data in bins along both x and y axes. 
    - If there are one or two points in a bin, take over those points
    - If there are more points in a bin, take over the minimum and maximum value
    - Appends the first and last point to make sure the same data range is used

    This allows keeping the original data in regions of low density but significantly 
    reducing the amount of data to plot in regions of high density. 
    At the same time, all the features are preserved with sufficiently dense binning.

    Parameters:
    - x: 1D array, input data for the x-axis.
    - y: 1D array, input data for the y-axis.
    - x_bins: 1D array, bins for the x-axis.
    - y_bins: 1D array, bins for the y-axis.

    Returns:
    - xfilt: 1D array, filtered x values.
    - yfilt: 1D array, filtered y values.
    
    Source of 1D script, modified here to account for 2D:
    https://stackoverflow.com/questions/53543782/how-can-i-subsample-an-array-according-to-its-density-remove-frequent-values
    
    """
    # Bin data along both x and y axes
    x_bins = np.digitize(x,x_bins)
    y_bins = np.digitize(y,y_bins)

    xfilt = []
    yfilt = []

    # Iterate over unique x and y bins
    for i in np.unique(x_bins):
        for j in np.unique(y_bins):
            # Select data points in the current x and y bin
            mask = (x_bins == i) & (y_bins == j)
            xi = x[mask]
            yi = y[mask]

            if len(xi) <= 2:
                xfilt.extend(list(xi))
                yfilt.extend(list(yi))
            else:
                xfilt.extend([xi[np.argmax(yi)], xi[np.argmin(yi)]])
                yfilt.extend([yi.max(), yi.min()])

    # Prepend/append first/last point if necessary
    if x[0] != xfilt[0]:
        xfilt = [x[0]] + xfilt
        yfilt = [y[0]] + yfilt
    if x[-1] != xfilt[-1]:
        xfilt.append(x[-1])
        yfilt.append(y[-1])

    # Sort the results based on x values
    sort = np.argsort(xfilt)
    return np.array(xfilt)[sort], np.array(yfilt)[sort]