# -*- coding: utf-8 -*-
"""
Created on Wed Jan 12 10:20:54 2022

@author: smvanessen

This script offers different functions for theoretical wave crest distribution shapes.
Rayleigh may also be used for wave-induced response peaks. 
The present implementation contains:
    - First-order narrow-banded Rayleigh
    - Second-order Forristall
    - Huang-Zhang for a single realisation (mean and 99% confidence bands)
    - Huang-Zhang for multiple realisations 

"""

# Required toolboxes and import syntaxes
import numpy as np
# from wavenumber import calc_wavenumber as wn

def make_cdf_Rayleigh(values,Hs):
    
    """ 
    % NOTE: FUNCTION CALLED CDF, BUT RETURNS 1-CDF (= exceedance probability)
    This function defines a Rayleigh CDF distribution
    
    INPUT ------------------ :
    values = Axis for which theoretical distribution is calculated. For the 
             amplitude distributions, these values are crest heights. For the 
             height distribution, these values are wave heights [m]
    Hs     = Significant wave height [m]
    
    OUTPUT ----------------- :
    cfd_Rayleigh_cr   = Cumulative distribution function Rayleigh for crest amplitudes
    cdf_Rayleigh_he   = Cumulative distribution function Rayleigh for crest heights
    
    REQUIRED TOOLBOXES ----- :
    import numpy as np
    
    IMPLEMENTATION BASED ON ----- :
    Scharnke, J. and Hennig, J. (2011). Statistical analysis of a set of basin 
    waves. 30th OMAE Conf., Rotterdam, The Netherlands.
    
    """
    
    cdf_Rayleigh_cr = np.exp(-8*(values**2/(Hs**2)))
    cdf_Rayleigh_he = np.exp(-2*(values**2/(Hs**2))) 

    return cdf_Rayleigh_cr, cdf_Rayleigh_he

def make_cdf_Forristall(values,Hs,T1,d,Vs):
    
    """ 
    % NOTE: FUNCTION CALLED CDF, BUT RETURNS 1-CDF (= exceedance probability)
    This function defines theoretical Forristall probability of exceedance 
    distributions based on the input x-values, for crests and heights of 
    second-order random waves. Assumes long-crested second-order waves.
    
    INPUT ------------------ :
    values    = Axis for which theoretical distribution is calculated. For the 
                amplitude distributions, these values are crest heights. For 
                the height distribution, these values are wave heights [m]
    Hs        = Significant wave height [m]
    T1        = Mean wave period (m0/m1) [s]
    d         = Water depth [m]
    Vs        = Forward speed or current speed [m/s] - used to calculate wave number

    OUTPUT ----------------- :
    cdf_Forristall_cr = Cumulative distribution function Forristall for crests
    cdf_Forristall_he = Cumulative distribution function Forristall for heights

    REQUIRED TOOLBOXES ----- :
    import numpy as np
    
    IMPLEMENTATION BASED ON ----- :
    -   Forristall, G.Z., 2000. Wave crest distributions: Observations and 
        second-order theory. Physical Oceanography 30, 1931–1943. 
        doi:10.1175/1520-0485(2000)030<1931:WCDOAS>2.0.CO;2.
    -   Scharnke, J. and Hennig, J. (2011). Statistical analysis of a set of 
        basin waves. 30th OMAE Conf., Rotterdam, The Netherlands.
    
    """
    
    # Gravitational acceleration [m/s^2]
    g = 9.81

    # Wave number (no current, deep water)
    om1 = 2*np.pi/T1
    k1 = wn(om1,d,Vs)[0] # The zero argument behind the function makes sure only the first output value is returned

    # Mean wave steepness S1
    S1 = 2*np.pi*Hs/(g*T1**2)

    # Ursell number Ur
    Ur = Hs/(k1**2*d**3)

    # Alpha and beta of the Weibul distribution that was fit by Forristall
    # for long-crested second-order waves.
    alpha = 0.3536  + 0.28925*S1 + 0.1060*Ur
    beta  = 2 - 2.15975*S1 + 0.0968*Ur**2

    # Crest distribution
    cdf_Forristall_cr = np.exp(-(values/(alpha*Hs))**beta)

    # Height distribution
    cdf_Forristall_he = np.exp(-2.263*(values/Hs)**2.126)

    return cdf_Forristall_cr,cdf_Forristall_he

def calc_cdf_HZ(coefs,S1,Ur,values,Hs):
    
    '''
    Help function for both Huang-Zhang distributions
    '''

    a0 = coefs[0]
    a1 = coefs[1]
    a2 = coefs[2]  
    a3 = coefs[3]  
    a4 = coefs[4]
    a5 = coefs[5]  
    a6 = coefs[6] 
    b0 = coefs[7]  
    b1 = coefs[8]  
    b2 = coefs[9]  
    b3 = coefs[10]  
    b4 = coefs[11]  
    b5 = coefs[12] 
    
    # Alpha and beta of the Weibul distribution that were fit by Huang-Zhang
    # alpha = a0 + a1*S1 + a2*S1^2 + a3*S1^3 + a4^4*S1^4 + a5*Ur + a6*Ur^2  # in paper, but a4^4 seems to be wrong.
    alpha = a0 + a1*S1 + a2*S1**2 + a3*S1**3 + a4*S1**4 + a5*Ur + a6*Ur**2 
    beta  = b0 + b1*S1 + b2*S1**2 + b3*S1**3 + b4*S1**4 + b5*Ur**2 
    
    # Define cumulative distribution
    cdf_HZ = np.exp(-(values/(alpha*Hs))**beta) 

    return cdf_HZ

def make_cdf_single3hrealisation_HuangZhang(values,Hs,T1,d,Vs):
    
    '''    
    Generate empirical Huang-Zhang probability of exceedance distributions 
    based on the input x-values, for crests and heights of non-linear waves. 
    Assumed long-crested non-linear waves. This distribution provides the
    mean and 99# upper and lower bands for a single 3h realisation (called
    PDSR in the publication). There is also a version for the distribution
    over a large number of 3h realisations (ensemble distribution, in another 
    script). NB it is assumed that Ur and S1 in the Huang-Zhang distribution
    are identical to the Ursell number Ur and the mean wave steepness S1 used 
    in for instance Forristall (2000). Each probability curve (mean, upper
    and lower 99#) have two regimes: probability < 10^-2 and probability >
    10^-2.
    
    INPUT:
    - values = axis for which theoretical distribution is calculated. For the
      crest distributions, these values are crest heights. For the heights
      distribution, these values are heights.
    - Hs = significant wave height [m]
    - T1 = mean wave period (m0/m1) [s]
    - d = water depth [m]
    - Vs = vessel speed [m/s]
    
    OUTPUT:
    - cdf_HZ_mean_cr = mean cumulative wave crest distribution function
                       Huang-Zhang for single 3h realisation 
    - cdf_HZ_low_cr  = lower limit 99# cumulative wave crest distribution 
                       function Huang-Zhang for single 3h realisation 
    - cdf_HZ_high_cr = upper limit 99# cumulative wave crest distribution 
                       function Huang-Zhang for single 3h realisation 
    
    Based on details in :
    Z. Huang and Y. Zhang (2018). SEMI-EMPIRICAL SINGLE REALIZATION AND 
    ENSEMBLE CREST DISTRIBUTIONS OF LONG-CREST NONLINEAR WAVES. 37th OMAE,
    Madrid, Spain.
    
    Original Matlab: SvE, Jul 2021.
    Matlab > Python: SvE, Jan 2022.
    '''
    
    # Gravitational acceleration [m/s^2]
    g = 9.81 
    
    # Wave number (current set equal to vessel speed)
    om1 = 2*np.pi/T1 
    k1 = wn(om1,d,Vs)[0]  # The zero argument behind the function makes sure only the first output value is returned

    # Mean wave steepness S1 (same as in Forristall, 2000)
    S1 = 2*np.pi*Hs/(g*T1**2) 
    
    # Ursell number Ur (same as in Forristall, 2000)
    Ur = Hs/(k1**2*d**3) 
    
    # Mean crest distribution single 3h realisation (probability < 10^-2)
    coef_mean_below = [0.2894, 12.3011, -662.632, 12153.3466, -68031.8045, 0.3779, -3.7904, 1.5277, 67.2118, -3683.1338, 63759.0846, -336712.3631, -8.1382] 
    cdf_HZ_mean_cr1 = calc_cdf_HZ(coef_mean_below,S1,Ur,values,Hs)

    # Lower 99% limit crest distribution single 3h realisation (probability < 10^-2)
    coef_down_below = [0.3752, 9.1269, -446.8393, 7837.572, -42682.1177, 0.8351, -5.7942, 2.4637, 82.5688, -4300.5362, 68857.4691, -337914.7489, 2.5661] 
    cdf_HZ_low_cr1 = calc_cdf_HZ(coef_down_below,S1,Ur,values,Hs)
    
    # Upper 99% limit crest distribution single 3h realisation (probability < 10^-2)
    coef_up_below = [0.1334, 13.0432, -751.0935, 14727.6571, -87711.081, 0.084, -5.2045, 0.7965, 44.9229, -2525.6956, 47184.1324, -270084.1075, -13.316] 
    cdf_HZ_high_cr1 = calc_cdf_HZ(coef_up_below,S1,Ur,values,Hs)

    # Mean crest distribution single 3h realisation (probability < 10^-2)
    coef_mean_above = [0.3712, 1.0087, -43.0667, 567.5292, -1173.1204, 0.1276, 0.3115, 2.006, 4.841, -321.181, 846.332, 27223.189, 0.832] 
    cdf_HZ_mean_cr2 = calc_cdf_HZ(coef_mean_above,S1,Ur,values,Hs)
    
    # Lower 99% limit crest distribution single 3h realisation (probability < 10^-2)
    coef_down_above = [0.3708, 0.9835, -39.9404, 598.3819, -2144.492, 0.2427, -0.5379, 2.212, 10.951, -637.673, 6707.056, -10244.141, 4.505] 
    cdf_HZ_low_cr2  = calc_cdf_HZ(coef_down_above,S1,Ur,values,Hs)
    
    # Upper 99% limit crest distribution single 3h realisation (probability < 10^-2)
    coef_up_above = [0.3516, 2.3892, -105.4167, 1557.3914, -5807.4512, 0.0791, 0.355, 1.62, 19.72, -909.665, 10465.556, -22952.443, -3.726] 
    cdf_HZ_high_cr2 = calc_cdf_HZ(coef_up_above,S1,Ur,values,Hs)
    
    ## Assemble full curves (different below and above probability = 10^-2)
    
    # Indices of probability level 10^-2
    i_lim1 = np.nonzero(cdf_HZ_mean_cr1 <= 10**(-2))[0][0]
    i_lim2 = np.nonzero(cdf_HZ_high_cr1 <= 10**(-2))[0][0]
    i_lim3 = np.nonzero(cdf_HZ_low_cr1 <= 10**(-2))[0][0]
    
    # Assemble total curves
    cdf_HZ_mean_cr = np.concatenate((cdf_HZ_mean_cr2[0:i_lim1],cdf_HZ_mean_cr1[i_lim1:]))
    cdf_HZ_high_cr = np.concatenate((cdf_HZ_high_cr2[0:i_lim2],cdf_HZ_high_cr1[i_lim2:]))
    cdf_HZ_low_cr = np.concatenate((cdf_HZ_low_cr2[0:i_lim3],cdf_HZ_low_cr1[i_lim3:]))
    
    return cdf_HZ_mean_cr,cdf_HZ_low_cr,cdf_HZ_high_cr
    

def make_cdf_multiple3hrealisations_HuangZhang(values,Hs,T1,d,Vs):
    
    '''
    Generate empirical Huang-Zhang probability of exceedance distributions 
    based on the input x-values, for crests and heights of non-linear waves. 
    Assumed long-crested non-linear waves. This distribution provides the
    distribution for a range of 3h realisations (called PDER in the 
    publication). There is also a version for the distribution of a single 3h 
    realisation (in another script). NB it is assumed that Ur and S1 in the 
    Huang-Zhang distribution are identical to the Ursell number Ur and the 
    mean wave steepness S1 used in for instance Forristall (2000). The 
    probability curve has two regimes: probability < 10^-2 and probability >
    10^-2.
    
    INPUT:
    - values = axis for which theoretical distribution is calculated. For the
      crest distributions, these values are crest heights. For the heights
      distribution, these values are heights.
    - Hs = significant wave height [m]
    - T1 = mean wave period (m0/m1) [s]
    - d = water depth [m]
    - Vs = vessel speed [m/s]
    
    OUTPUT:
    - cdf_HZ_cr  =  cumulative wave crest distribution function Huang-Zhang 
                    for single 3h realisation
    
    Based on details in :
    Z. Huang and Y. Zhang (2018). SEMI-EMPIRICAL SINGLE REALIZATION AND 
    ENSEMBLE CREST DISTRIBUTIONS OF LONG-CREST NONLINEAR WAVES. 37th OMAE,
    Madrid, Spain.
    
    Original Matlab: SvE, Jul 2021.
    Matlab > Python: SvE, Jan 2022.
    '''
    
    # Gravitational acceleration [m/s^2]
    g = 9.81 
    
    # Wave number (current set equal to vessel speed)
    om1 = 2*np.pi/T1 
    k1 = wn(om1,d,Vs)[0]  # The zero argument behind the function makes sure only the first output value is returned

    # Mean wave steepness S1 (same as in Forristall, 2000)
    S1 = 2*np.pi*Hs/(g*T1**2) 
    
    # Ursell number Ur (same as in Forristall, 2000)
    Ur = Hs/(k1**2*d**3) 
    
    # Mean crest distribution single 3h realisation (probability < 10^-2)
    coef_below = [0.3242, 11.7467, -652.147, 12308.3001, -70529.2504, 0.3785, -3.8837, 1.7321, 72.7179, -4093.6916, 72132.2957, -386504.7502, -9.9594]  
    cdf_HZ_cr1 = calc_cdf_HZ(coef_below,S1,Ur,values,Hs)
    
    # Mean crest distribution single 3h realisation (probability < 10^-2)
    coef_above = [0.3733, 0.9398, -40.0095, 512.0601, -849.0734, 0.1294, 0.2882, 2.0411, 3.6068, -272.2806, -58.9649, 32786.2976, 0.9003] 
    cdf_HZ_cr2 = calc_cdf_HZ(coef_above,S1,Ur,values,Hs)
    
    ## Assemble full curve (different below and above probability = 10^-2)
    
    # Index of probability level 10^-2
    i_lim1 = np.nonzero(cdf_HZ_cr1 <= 10**(-2))[0][0]
    
    # Assemble total curve
    cdf_HZ_cr = np.concatenate((cdf_HZ_cr2[0:i_lim1],cdf_HZ_cr1[i_lim1:]))

    return cdf_HZ_cr

def wn(om,d,U):
    
    """ 
    This function determines the wave number k based on the radial wave frequency, 
    the depth and the current velocity (or vessel speed, neglecting wave-current 
    interaction effects and only accounting for the Doppler encounter frequency 
    effect). This is the iterative solution of the linear dispersion relation:  
        k*g*tanh(k*h)=(om-k*U)^2
    A solution is found numerically with Newton's method. The first guess is
    the infinite depth wave number.
    
    INPUT ------------------ :
    om   = Frequency axis [rad/s]
    d    = Water depth [m]
    U    = Current speed parallel to the waves [m/s]. Can also be used for forward speed.
    
    OUTPUT ----------------- :
    k    = Wave number [-]
    c_ph = Wave phase velocity [m/s]
    c_gr = Wave group velocity [m/s]
    
    REQUIRED TOOLBOXES ----- :
    import numpy as np
    
    BASED ON MATLAB SCRIPT REVISIONS ----- :
    10 jan 2007 RvDijk:             Check for om=0 & only calc blockage if U <>0
    23 sep 2007 Arjan Voogt:        Check current blockage (line 21)
    2 feb 2008 Arjan Voogt:         Update warning message
    8 feb 2008 Arjan Voogt:         Removed critial error in iteration procedure. 
                                    The original Newton method can find the solution 
                                    of the dispersion relation with negative wavenumber 
                                    k. Therefore k_new=k_old-func./afg is changed to 
                                    k_new=abs(k_old-func./afg)
    12 mar 2021 Sanne van Essen:    Added some comments, replaced disp(sprintf()) with fprintf().
    8 jun 2021 Sanne van Essen:     Restructured
    12 jan 2022 Sanne van Essen:    To Python
    
    """
    
    # Default iteration settings
    maxerror = 0.0000001
    maxiter = 50
    
    # # Check size
    # om = om(:)

    # Gravitational acceleration [m/s^2]
    g = 9.81
    
    # Check whether current blockage occurs
    kk = np.arange(0,100.01,0.01)
    maxom = np.max(np.sqrt(kk*g*np.tanh(kk*d))+kk*U)
    if (np.max(om) > maxom) and (U != 0):
        blocked = np.size(om[om > maxom])
        om = om[om <= maxom]
        print('Warning: blockage occurs for om > %0.2f rad/s at waterdepth %0.1f m and current velocity %0.1f m/s' % (maxom,d,U))
    else:
        blocked = 0
    
    # Check for om = 0 (only if om is not a scalar)
    if isinstance(om,list) and om[0] == 0:
        om = om[1:] # using index -1 for the end removes one index somehow..
        k0 = 0
    else:
        k0 = []
   
    # First guess
    k_old = om**2/g
    error = 2*maxerror
    i = 0

    # Update this guess iteratively
    while (np.max(error) > maxerror) and (i < maxiter):
        i = i+1
        func = (om - k_old*U)**2 - g*k_old*np.tanh(k_old*d)
        afg = -2*U*om + 2*k_old*U**2 - g*np.tanh(k_old*d) - g*k_old*d/np.cosh(k_old*d)**2
        k_new = np.abs(k_old - func/afg)
        error = np.abs(k_old - k_new)/((om - k_old*U)**2/9.81)
        k_old = k_new
    
    # Concatenate with blocked part (only if om is not a scalar)
    if isinstance(om,list):
        temp = np.max(k_old)*np.ones(blocked)
        k = np.concatenate((k_old,temp),axis=0)
    else:
        k = k_old
    
    # Show warning
    if i == maxiter:
        print('Warning: maximum number of iterations exceeded at frequencies [rad/s]:')
        ind = np.where(error > maxerror)
        print(om[ind])
        print('For depth %0.1f m and current velocity %0.1f m/s' % (d,U))
    
    # Phase and group velocity
    c_ph = om/k
    c_gr = c_ph/2*(1+2*k*d/np.sinh(2*k*d))+U

    # Add elements if the first entry is zero
    if k0 == 0:
        k = np.append(k0,k)
        om = np.append(0,om)
        c_ph = np.append(np.nan,c_ph)
        c_gr = np.append(np.nan,c_gr)
    
    return k,c_ph,c_gr