# -*- coding: utf-8 -*-
"""
Created on Mon 9 Jan 2022

@author: smvanessen

This script offers different wave spectral shapes. Present inplementation contains:
    - JONSWAP frequency spectrum
    - cos-2s directional spectrum

"""

# Required toolboxes and import syntaxes
import numpy as np

def make_spec_jons(Hs,Tp,gamma,om):
    
    """ 
    This function defines a JONSWAP spectrum
    
    INPUT ------------------ :
    Hs    Significant wave height [m]
    Tp    Peak wave period [s]
    gamma Peak enhancement factor of JONSWAP spectrum (3.3 = standard JONS, 1.0 = Pierson-Moskowitz fully developed)
    om    Frequency axis, 1D array [rad/s]. May start at zero or not.
    
    OUTPUT ----------------- :
    S     JONSWAP wave spectral shape based on frequency axis [m^2.s/rad]
    
    REQUIRED TOOLBOXES ----- :
    import numpy as np
    
    """
    
    # If first frequency is zero, remove for now (to avoid trouble in calculation of 'S_hz')
    if om[0] == 0:
        # print('first element of frequency axis is zero')
        om1 = om[1:]
    else:
        # print('first element of frequency axis is not zero')
        om1 = om
    
    # Input peak wave frequency fp and frequency axis f in [Hz] instead of [rad/s]
    fp = 1/Tp
    f = om1/(2*np.pi)
    
    # 1D Tau parameter array (first make zeros array with the same size as f)
    tau = np.zeros((len(f)))
    tau[f<=fp] = 0.07
    tau[f>fp] = 0.09
    
    # Alpha* parameter
    alst = 0.0624/(0.230+0.0336*gamma-0.185*(1.9+gamma)**(-1))
    
    # Spectral form ('approximate form' from ITTC 2002 procedure)
    term = np.exp(-(f-fp)**2/(2*tau**2*fp**2))
    S_hz = 2*np.pi*alst*Hs**2*(f**(-5)/fp**(-4))*np.exp(-(5/4)*(f/fp)**(-4))*gamma**(term)
    
    # Transform to radial frequency axis
    S = S_hz/(2*np.pi)**2
    
    # Add zero if first input frequency was zero 
    if om[0] == 0:
        S= np.concatenate(([0],S))
    
    return S


def make_dirspec_cos2s(s,ddom,dirs):
    
    '''
    This function defines a cosine-2s wave directional spreading function
   
    INPUT ------------------ :
    s         Spreading parameter s [-]
    ddom      Dominant direction [rad] 
    dirs      Directional axis [rad]  
    
    OUTPUT ----------------- :
    D         Cos-2s wave directional spreading function based on directional axis [1/rad]
    sig_d     Directional width [rad]
   
    REQUIRED TOOLBOXES ----- :
    import numpy as np
   
    '''
    
    # Cosine 2s spreading function
    D1 = np.cos((dirs - ddom)/2)**(2*s)
    
    # Normalise to integrated value 1
    # ddirs = (dirs(end) - dirs(1))/(len(dirs) - 1)
    ddirs = (dirs[-1] - dirs[0])/(len(dirs) - 1)
    D = 1/(np.sum(D1)*ddirs)*D1

    # Directional width
    sig_d = np.sqrt(2/(s+1))
    
    return D,sig_d




