# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 13:43:17 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Import toolboxes
import numpy as np
import random
import matplotlib.pyplot as plt
import os

# Import functions
from functions.distribution_scripts import calc_distribution,MPM_from_DNR,MPM_from_DEM
# from functions.distribution_scripts import fit_weibull3_lsq
from functions.subsample_routines import subsample_binned

# Close all figures
plt.close('all')


#############
# START INPUT
#############

# Total number of available seeds (have to be generated already by the script
# MF_GPR_Lin2ndOrd_NseedsX_finalpaper.py
Ns = 50

# Number of picked seeds for the bootstrapping procedure. Set to 5 in the paper.
M = 5

# Number of seed picking realisations (Y times M seeds out of the available Ns)
Y = 500 # paper
# Y = 10 # debugging

# Folder with files to read (rel. to present folder).. These files are written 
# to this file by running Adaptive Screening with MF-GPR and sampling option B.
in_filefolder = 'waveinputfiles'

# Folder in which to store output figure (rel. to present folder).
out_figfolder = 'out_figs'

# Prediction range for Weibull fit
v_range = np.arange(7,15,0.01)

# Target duration (here fixed at 1h)
dur = 3600

# Peak wave period [s]
Tp = 11

# Figure size
fsize_h = 7
fsize_v = 5

#############
# END INPUT
#############


# Check wether output directories exist, if not make them
isExist1 = os.path.exists(out_figfolder)
if not isExist1:
   os.makedirs(out_figfolder)
  
# Read peaks from CSV (concatenated over all <Ns> seeds). We read GoodInd data, 
# as these are the only relevant data for the conventional procedure.
AllTrueDatSec = np.loadtxt(f'{in_filefolder}/Lin2nd_trueLFandHFdata_GoodInd_{Ns}seeds_true2nd.csv', delimiter=',')
t_peaks_2nd_all = AllTrueDatSec[:,0]
peaks_2nd_all= AllTrueDatSec[:,1]

# Read extrema from CSV (1 value per seed). We read GoodInd data, as these are 
# the only relevant data for the conventional procedure.
AllTrueEM = np.loadtxt(f'{in_filefolder}/Lin2nd_trueLFandHFdataEM_GoodInd_{Ns}seeds.csv', delimiter=',')
EM_lin_all = AllTrueEM[:,0]
EM_2nd_all = AllTrueEM[:,1]

print(f'Reading CSV with LF and HF MCS results over {Ns} seeds done')

# Average number of peaks in target duration
ne_exp = dur/Tp

# Exposure duration probability level in the DNR
pexp = 1/ne_exp

# Rough duration per seed (disregarding last part of last seed without peaks)
tmax = t_peaks_2nd_all[-1]
t_per_seed = np.floor(tmax/Ns)

# Initialise
t_st = np.zeros(Ns)
t_en = np.zeros(Ns)
MPM_weib = np.zeros(Ns)

# Initialise
MPM_weib_DEM = np.zeros(Y)
MPM_weib_DEM_fit = np.zeros(Y)

# Bootstrapping over Y seed picking realisations (every time pick M seeds out of the available Ns)
for jj in range(Y):
    
    # Create a list of Ns options
    options = list(range(1, Ns))
    
    # Select M random elements from the list of options
    selected = random.sample(options, M)
    
    # Print the selected elements
    print(f"Loop {jj}. Selected {M} indices out of 0 to {Ns-1} options in seed picking realisation {jj+1}: {selected}")
    
    # Initialise
    DEM = np.zeros(M)
    DEM_fit = np.zeros(M)
    
    # Prepare figure (do this only for first 10 seed picking realisations)
    if jj < 10:
        plt.figure(100+jj,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        
        plt.figure(200+jj,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
    
    # Loop over selected seeds in seed picking realisation jj
    for num,ii in enumerate(selected):
        
        # print(f'{num},{ii}')
        
        
        # Split the time traces in <Ns> seeds again, select start and end times for present seed
        t_st[ii] = t_per_seed*ii
        if ii != Ns-1:
            t_en[ii] = t_per_seed*ii + t_per_seed - 0.001
        else:
            # Last seed is slightly longer
            t_en[ii] = tmax
            
        i_st = np.argmax(t_peaks_2nd_all>=t_st[ii])
        i_en = np.argmax(t_peaks_2nd_all>=t_en[ii])
        print(f'Seed {ii+1}, time limits {t_st[ii]}-{t_en[ii]}s')
        
        # HF peaks within these time bounds
        peaks_2nd_all_seed = peaks_2nd_all[i_st:i_en]
        
        # Make exceedance distribution
        v_seed,p_seed = calc_distribution(peaks_2nd_all_seed)
            
        # # For each seed, fit 3-param Weibull to all peaks
        # topPerc = 30
        # maxit = 100
        # _,_,theta,alpha,beta,_ = fit_weibull3_lsq(p_seed,v_seed,topPerc)
        # p_weib_exc = np.exp(-((v_range - theta)/alpha)**beta);
        
        # Add to plot (do this only for first 10 seed picking realisations)
        if jj < 10:
            
            # Subsample results
            binsize = 0.05
            bins = np.arange(0,15,binsize)
            v_seed_sampled,p_seed_sampled = subsample_binned(v_seed,p_seed,bins)
            # v_range_sampled,p_weib_exc_sampled = subsample_binned(v_range,p_weib_exc,bins)

            # # With fit
            # plt.figure(100+jj)
            # if ii == selected[0]:
            #     # plt.plot(v_seed,p_seed,'-r',label='Measured HF peaks per seed')
            #     plt.plot(v_seed_sampled,p_seed_sampled,color='r', marker='d',markersize=5,linestyle='-',label=f'All HF peaks in {M} seeds')
            #     plt.plot(v_range_sampled,p_weib_exc_sampled,'-.',color='purple',label=f'Weibull fits to top {topPerc}$\%$ of these peaks')
            # else:
            #     # plt.plot(v_seed,p_seed,'-r')
            #     plt.plot(v_seed_sampled,p_seed_sampled,color='r', marker='d',markersize=5,linestyle='-')
            #     plt.plot(v_range_sampled,p_weib_exc_sampled,'-.',color='purple')
            
            # Without fit
            plt.figure(200+jj)
            if ii == selected[0]:
                plt.plot(v_seed_sampled,p_seed_sampled,color='r', marker='d',markersize=5,linestyle='-',label=f'All HF peaks in {M} seeds')
            else:
                plt.plot(v_seed_sampled,p_seed_sampled,color='r', marker='d',markersize=5,linestyle='-')
             
        
        # Save value at Pexp per seed, without and with Weibull fit. Use interpolation to avoid the issue of the 
        DEM[num] = MPM_from_DNR(v_seed,p_seed,ne_exp) # Use this per seed for the interpolation
        # DEM_fit[num] = MPM_from_DNR(v_range,p_weib_exc,ne_exp) # Use this per seed for the interpolation

    # Make exceedance distribution of the DEM for the present seed picking realisation, without and with Weibull fit
    v_DEM,p_DEM = calc_distribution(DEM)
    # v_DEM_fit,p_DEM_fit = calc_distribution(DEM_fit)

    # Derive MPM from fit per seed picking realisation, based on Weibull DEM over M seeds, without and with Weibull fit
    MPM_weib_DEM[jj] = MPM_from_DEM(v_DEM,p_DEM) # Double-check!
    # MPM_weib_DEM_fit[jj] = MPM_from_DEM(v_DEM_fit,p_DEM_fit)
    

    # Finalise figures (do this only for first 10 seed picking realisations)
    if jj < 10:
        # plt.figure(100+jj)
        # plt.plot([0,max(v_range)],[pexp,pexp],':',color='black') #,label='$p_{exp}$')
        # plt.text(14,pexp*1.4,'$P_{exp}$')
        # plt.legend()
        # plt.xlim([0,15])
        # plt.ylim([10**(-5),1])
        # plt.ylabel('P(C\'\'$\geq$c) [-]')
        # plt.xlabel('c [m]')
        # plt.savefig(f'{out_figfolder}/Lin2nd_ConventionalDistr_{M}seeds_seedpickreal{jj}_inclfit_upd.pdf',bbox_inches='tight')
        
        plt.figure(200+jj)
        plt.plot([0,max(v_range)],[pexp,pexp],':',color='black') #,label='$p_{exp}$')
        plt.text(14,pexp*1.4,'$P_{exp}$')
        plt.plot(DEM,pexp*np.ones(np.shape(DEM)),'xr',markersize=10,label='Ensemble maxima')
        plt.plot(MPM_weib_DEM[0],pexp,'ob',markersize=10,markerfacecolor='none',label=f'MPM over these {M} seeds')
        plt.legend()
        plt.xlim([0,15])
        plt.ylim([10**(-5),1])
        plt.ylabel('P(C\'\'$\geq$c) [-]')
        plt.xlabel('c [m]')
        plt.savefig(f'{out_figfolder}/Fig9_Lin2nd_ConventionalDistr_{M}seeds_seedpickreal{jj}_upd.pdf',bbox_inches='tight')
       
# # Mean and U95 over the found DEM-based MPM values, after fitting
# Mean_MPM_fit = np.mean(MPM_weib_DEM_fit)
# U95_MPM_fit = 1.96*np.std(MPM_weib_DEM_fit)
# print(f'With Weibull fitting: Mean MPM value conventional method with {M} seeds at {dur/3600}h: {Mean_MPM_fit} m with U95 {U95_MPM_fit} m')

# Mean and U95 over the found DEM-based MPM values
Mean_MPM = np.mean(MPM_weib_DEM)
U95_MPM = 1.96*np.std(MPM_weib_DEM)
print(f'Without fitting: Mean MPM value conventional method with {M} seeds at {dur/3600}h: {Mean_MPM} m with U95 {U95_MPM} m')

